
from __future__ import print_function, division

"""
Reciprocal Space Volume

This class is just for holding a volume and loading/saving it to disk
Another class (rsv_mapper) will add images into it.
"""


# ImageD11_v1.0 Software for beamline ID11
# Copyright (C) 2005-2007  Jon Wright
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  0211-1307  USA


import logging, numpy, h5py

class rsv(object):
    """
    A reciprocal space volume
    """
    def __init__(self, dimensions, bounds, np, **kwds ):
        """
        dimensions = NX*NY*NZ grid for the space
        uspace = a 3x3 matrix describing the grid
         eg:
            pixel at vol[i,j,k] comes from:
               [i,j,k] = (uspace).gvec
          gvec is the scattering vector in reciprocal space
          so uspace are vectors in real space 
        uorigin = a 3 vector giving the position of the [0,0,0] pixel
        """
        assert len(dimensions) == 3
        self.SIG = None # signal
        self.MON = None # monitor
        self.NR = [int(x) for x in dimensions] # dimensions
        self.NORMED = None
        self.bounds = bounds  # boundary in reciprocal space
        self.np = np          # px per hkl
        self.metadata = kwds
        #  Do not allocate in constructor for now - make caller care about
        # memory usage
        #  self.allocate_vol()

    def allocate_vol( self ):
        """ 
        Allocates memory for a volume data
        """
        if self.NR is None:
            raise Exception("Cannot allocate rsv")
        total = int(self.NR[0]*self.NR[1]*self.NR[2])
        print("rsv: memory used = %.2f MB"%(total*8.0/1024/1024))
        print("dim: %d %d %d"%(self.NR[0],self.NR[1],self.NR[2]))
        self.SIG = numpy.zeros( total, numpy.float32 )
        self.MON = numpy.zeros( total, numpy.float32 )
        
        
    def normalise( self , savespace = True):
        """
        Return the normalised but avoid divide by zero
        """
        if savespace:
            self.NORMED = self.SIG
        else:
            self.NORMED = numpy.zeros( self.SIG.shape, numpy.float32 )
        import sys
        print(self.NORMED.shape[0])
        for i in range( self.NORMED.shape[0] ):
            print(i,"\r", end=' ')
            sys.stdout.flush()
            msk = (self.MON[i] < 0.1).astype(numpy.uint8)
            numpy.add( self.MON[i], msk, self.MON[i])
            numpy.divide( self.SIG[i],
                          self.MON[i],   # divide by mon + 1
                          self.NORMED[i] )
            numpy.subtract( self.MON[i], msk, self.MON[i])
            numpy.subtract( 1, msk, msk )
            numpy.multiply( self.NORMED[i], msk, self.NORMED[i] )
        print(i)

            
    plnames = {
        0 :0, "h":0, "H":0,
        1 :1, "k":1, "K":1,
        2 :2, "l":2, "L":2,
        }
       

    def slice(self, plane, num):
        """
        return signal on plane index num 
        """
        if plane not in self.plnames:
            raise Exception("Plane should be one of %s"%(
                        str(self.plnames)))
        p = self.plnames[plane]
        # floor(x+0.5) is nearest integer
        ind = int(numpy.floor( num * self.np + 0.5) - self.bounds[p][0])
        # convert this back to num
        testnum = 1.0*(self.bounds[p][0] + ind )/self.np
        if abs(testnum - num)>1e-6:
            logging.info("Nearest plane to %f is %f"%(num, testnum))
        if ind < 0 or ind >= self.NR[p]:
            print(ind,num,self.np,self.bounds)
            raise Exception("slice is out of volume bounds")
        if self.NORMED is None:
            self.normalise()
        if len(self.NORMED.shape) == 1:
            self.NORMED.reshape(self.NR)
        if p==0:
            return self.NORMED[ind, :, :]
        if p==1:
            return self.NORMED[:, ind, :]
        if p==2:
            return self.NORMED[:, :, ind]


def getbounds( vol, plane ):
    """
    Returns the extent argument to use for pylab.imshow when plotting
    a plane
    """
    inds = [0,1,2]
    inds.remove( vol.plnames[plane] )
    imin = vol.bounds[inds[0]][0]*1.0/vol.np
    imax = vol.bounds[inds[0]][1]*1.0/vol.np
    jmin = vol.bounds[inds[1]][0]*1.0/vol.np
    jmax = vol.bounds[inds[1]][1]*1.0/vol.np
    # left, right, top, bottom
    return jmin,jmax,imax,imin


def writevol(vol, filename):
    """
    Write volume in vol to filename
    
    Compress -1 is for the zeros, which there might be a lot of
    """
    if not isinstance( vol, rsv ):
        raise Exception("First arg to writevol should be an rsv object")
    for a in [vol.NR, vol.SIG, vol.MON]:
        if a is None:
            raise Exception("Cannot save rsv, has not data in it")
    volout = h5py.File( filename,"w")
    if vol.SIG.dtype != numpy.float32:
        logging.warning("rsv SIG was not float32, converting")
        vol.SIG = vol.SIG.astype(numpy.float32)
    volout.create_dataset( "signal",
                           (vol.NR[0],vol.NR[1],vol.NR[2]),
                           vol.SIG.dtype,
                           data = vol.SIG,
                           compression = 'gzip', 
                           compression_opts = 1)
    if vol.MON.dtype != numpy.float32:
        logging.warning("rsv MON was not float32, converting")
        vol.MON = vol.MON.astype(numpy.float32)
    volout.create_dataset( "monitor",
                           (vol.NR[0],vol.NR[1],vol.NR[2]),
                           vol.MON.dtype,
                           data = vol.MON,
                           compression = 'gzip',
                           compression_opts = 1)
    volout.attrs['bounds'] = vol.bounds
    volout.attrs['np'] = vol.np
    for key, value in vol.metadata.items():
        volout.attrs[key]=value
    volout.flush()
    volout.close()

    
def writenormedvol(vol, filename):
    """
    Write volume in vol to filename - save only the normalised
    to avoid using so much memory
    
    Compress -1 is for the zeros, which there might be a lot of
    """
    if not isinstance( vol, rsv ):
        raise Exception("First arg to writevol should be an rsv object")
    
    for a in [vol.NR, vol.NORMED]:
        if a is None:
            raise Exception("Cannot save rsv, has not data in it")
    volout = h5py.File( filename,"w")
    if vol.NORMED.dtype != numpy.float32:
        logging.warning("rsv NORMED was not float32, converting")
        vol.NORMED = vol.NORMED.astype(numpy.float32)
    volout.create_dataset( "signal",
                           (vol.NR[0],vol.NR[1],vol.NR[2]),
                           vol.NORMED.dtype,
                           data = vol.NORMED,
                           compression = 'gzip', 
                           compression_opts = 1)
    volout.attrs['bounds'] = vol.bounds
    volout.attrs['np'] = vol.np
    for key, value in vol.metadata.items():
        volout.attrs[key]=value
    volout.flush()
    volout.close()


def mem():
    """ debug the memory usage """
    import os
    os.system('ps v -p %s'%(os.getpid()))

def readvol(filename, savespace=False ):
    """
    Read volume from a file
    returns an rsv object
    Take care to allocate and read to avoid temporaries
    """
    
    volfile = h5py.File(filename)
    if not 'signal' in list(volfile.keys()):#listnames():
        raise Exception("Your file %s is not an rsv"%(filename))
    sig = volfile['signal']
    bounds = volfile.attrs['bounds']
    np = volfile.attrs['np']
    vol = rsv( sig.shape, bounds, np )
    # allocate array empty
    #mem()
    if savespace:
        vol.SIG = sig
    else:
        vol.SIG = numpy.empty( sig.shape, sig.dtype )
        #mem()
        sig.read_direct( vol.SIG )
    #mem()
    for name, value in volfile.attrs.items():
        vol.metadata[name] = value
    #mem()
    if 'monitor' in list(volfile.keys()):#listnames():
        mon = volfile['monitor']
        assert mon.shape == vol.SIG.shape
        if savespace:
            vol.MON = mon
        else:
            vol.MON= numpy.empty( mon.shape, mon.dtype)
            mon.read_direct( vol.MON )
    else:
        vol.MON = None
        vol.NORMED = vol.SIG
    #mem()
    if savespace:
        vol.hdf_file_object = volfile
    else:
        volfile.close()
    #mem()
    return vol








from __future__ import print_function




# ImageD11 Software for beamline ID11
# Copyright (C) 2008  Jon Wright
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  0211-1307  USA

"""
If you google for "fit plane points" you can find similar things. This 
uses an SVD to fit the plane to the points. The story goes:

    center of mass is in the plane
    vectors from the c.o.m to the points are in a space
    svd finds principal vectors for that space
    null vector (or smallest) is perpendicular to plane holding points

Then plane computations go via n.(r-c)=0

Enjoy!
"""


# Fit a least squares plane to some data
# Data are likely to be the edge of a box from an integration
# ... but could be anything (eg: g-vectors etc)

import numpy as np

class plane:

    def __init__(self, normal, point):
        """
        Normal - vector normal to the surface
        point  - a point on the surface

        Equation of plane is normal.(x - point) = 0
        """
        assert len(point) == len(normal) == 3, "3D space please"
        assert abs (np.dot(normal , normal) - 1) < 1e-6, \
                "Normalise normal please"
        self.point = point
        self.normal = normal

    def calc(self, points, index=2):
        """
        points - [x, y, z]
        index  - compute x or y or z
        """
        assert index in [0,1,2],"3D space please"
        assert isinstance(points, np.ndarray),"Please use numpy array for points"
        assert points.shape[1] == 3
        assert abs(self.normal[index]) > 1e-6, "Sorry, your plane is parallel"
        points[:, index ] = 0.
        # Compute vectors from self.point to points
        vecs = points - self.point
        # Projections along normal
        proj = np.dot(self.normal, vecs.T)
        # Now determine points[index, :] such that proj == 0 everywhere
        points[:, index ] = -proj/self.normal[index]
        return points



def fitplane(x, y, z):
    """
    x, y and z are the co-ordinates of the points to fit to
    """
    assert len(x) == len(y) == len(z)
    cx = np.mean( x ) 
    cy = np.mean( y ) 
    cz = np.mean( z ) 
    skinny = np.array( [np.asarray(x) - cx ,
                        np.asarray(y) - cy ,
                        np.asarray(z) - cz ] )     
    u, s, vh = np.linalg.svd( skinny , full_matrices = 0 )
    # print "Through",cx,cy,cz,"Normal is u",u[:,-1]
    return plane( u[:,-1].copy(), np.array([cx, cy, cz]))


def test():
    p = fitplane( [ 1 , 0 , 0 ]    ,
                  [ 0 , 1 , 0 ]    ,
                  [ 0 , 0 , 1 ]    )
    for i in range(3):
        assert abs(abs(p.normal[i]) - 1/np.sqrt(3))< 1e-6
    assert np.allclose( p.normal[0], p.normal[1])
    assert np.allclose( p.normal[0], p.normal[2])

    assert (p.calc( np.array([[ -1, 0, 0]] ) ) == np.array( [-1, 0, 2] )).all()
    assert (p.calc( np.array([[  2, 0, 0]] ) ) == np.array( [ 2, 0,-1] )).all()
    assert (p.calc( np.array([[  0, 2, 0]] ) ) == np.array( [ 0, 2,-1] )).all()
    assert (p.calc( np.array([[  0,-1, 0]] ) ) == np.array( [ 0,-1, 2] )).all()
    
    print("Tests look OK")

if __name__ == "__main__":
    test()





from __future__ import division, print_function
import numpy as np, time
from ImageD11.cImageD11 import localmaxlabel
import unittest, sys


def localmax( im, out=None ):
    """ reference implementation 
    ... very slow 
    """
    neighbours = [ (-1,-1),(-1,0), (-1,1),
                   (0,-1), (0,0),   (0,1),
                   (1,-1), (1,0),  (1,1) ]
    out = np.arange( im.shape[0]*im.shape[1], dtype=np.int )
    out.shape = im.shape
    out0 = out.copy()    # pointers
#    print("running python localmax", im.shape,end=" " )
    for i in range(1,im.shape[0]-1):
        for j in range(1,im.shape[1]-1):
            mx = -1e10
            for k in range(i-1,i+2):
                for l in range(j-1,j+2):
                    try:
                        if im[k,l] > mx:
                            out[i,j] = out0[k,l]
                            mx = im[k,l]
                    except:
                        print(i,j,k,l,im.shape)
                        raise
            # now out addresses a neighbour
 #   print("1", end=" " )
    # Maximal pixels
    maxpx = (out==out0)
    # Final labels
    npk = 0
    labels = np.zeros( im.shape, np.int32 )-1
    labels[0]=0   # borders
    labels[-1]=0
    labels[:,0]=0
    labels[:,-1]=0
  #  print("2", end=" " )
    for i in range(1,im.shape[0]-1):
        for j in range(1,im.shape[1]-1):
            if maxpx[i,j]:
                npk += 1
                labels[i,j]=npk
                out[i,j]=0
  #  print("3", end=" " )
  #  sys.stdout.flush()
    # Now relabel each pixel to point to the max...
    pon=False
    for i in range(1,im.shape[0]-1):
        for j in range(1,im.shape[1]-1):
            adr = out[i,j]
            if adr == 0:
                continue
            while 1:
                adrnew = out.flat[adr]
                if adrnew == 0 or labels.flat[adr]==0:
                    break
                adr = adrnew
            l = labels.flat[adr] # take from peak
            labels[i,j] = l
            adr = out[i,j]
            while 1:
                labels.flat[adr] = l
                adrnew = out.flat[adr]
                if adrnew == 0:
                    break
                out.flat[adr] = 0
                adr = adrnew
   # print(".", end=" " )                            
   # print ("ok")
    return labels


def make_test_image( N=256, w=3 ):
    im = np.zeros( (N, N), np.float32 )
    # patch size to hold peaks
    pN = w*5+1  
    pk = np.zeros( (pN,pN), np.float32 )
    pi,pj = np.mgrid[ 0:pN, 0:pN ]
    # two peaks, separated by i
    ni = N//pN
    s = 5.
    for i in range(0, ni):
        # at max separation we want 0 - pk - pk - pN
        # so:
        dj = (pN/3.0)*(i/(ni-1))
        di =  dj / 5.

        ci1 = pN//2  + di
        ci2 = pN//2  - di
        cj1 = pN//2  + dj
        cj2 = pN//2  - dj
        p1 = np.exp( - (( pi - ci1 )**2 + (pj - cj1)**2)/s)
        p2 = np.exp( - (( pi - ci2 )**2 + (pj - cj2)**2)/s)
        for j in range(0, N//pN):
            im[ i*pN:(i+1)*pN, j*pN:(j+1)*pN ] = p1 + p2*(j+1)/ni
    return im

def demo():
    import pylab as pl
    im = make_test_image()
    l1 = localmax(im)
    l2 = np.zeros(im.shape,np.int32)
    wk = np.zeros(im.shape,np.int8)
    npks = localmaxlabel(im, l2 , wk, cpu=0 )
    ax1 = pl.subplot(2,2,1)
    pl.imshow(im)
    pl.xlim( 0,im.shape[0])
    pl.ylim( 0,im.shape[1])
    ax1 = pl.subplot(2,2,2, sharex=ax1, sharey=ax1)
    pl.imshow((l1*13.0)%37)
    pl.xlim( 0,im.shape[0])
    pl.ylim( 0,im.shape[1])
    ax1 = pl.subplot(2,2,3, sharex=ax1, sharey=ax1)
    pl.imshow((l2*13.0)%37)
    pl.xlim( 0,im.shape[0])
    pl.ylim( 0,im.shape[1])
    pl.show()

class test_localmaxlabel( unittest.TestCase ):
    def setUp( self ):
        self.im = make_test_image(128)[:107,:109].copy()
        self.dopy = True
    def test1( self):
        im = self.im
        if self.dopy:
            start = time.time()
            l1 = localmax(im)
            end = time.time()
            pytime = (end-start)*1000.
        l = [l1,]
        wk = [0,]
        times = [pytime,]
        names = ["python",]
        for cpu, name in [(0,"ansi"),(1,"sse2"),(2,"avx2")]:
            li = np.zeros(im.shape,np.int32)
            wi = np.zeros(im.shape,np.int8)
            start = time.time()
    #        print("cpu",cpu)
            npks = localmaxlabel(im.copy(), li , wi, cpu=cpu )
            end=time.time()
            ctime = (end-start)*1000.
            l.append( li )
            times.append( ctime )
            names.append( name )
        lref = l[0]
        print("Timing",end=" ")
        for i in range(len(names)):
            print( "%s %.3f ms"%( names[i], times[i]), end=" ")
        print()
        for li in l[1:]:
            self.assertTrue((lref == l1).all())
    def test3( self):
     #   print("in test 2")
        im = make_test_image(N=2048)
        l1 = np.zeros(im.shape,np.int32)
        l2 = np.zeros(im.shape,np.int32)
        l3 = np.zeros(im.shape,np.int32)
        wk = np.zeros(im.shape,np.int8)
        start = time.time()
        for i in range(10):
            npks = localmaxlabel(im, l1 , wk )
        end=time.time()
        npks = localmaxlabel(im, l1 , wk, 10 )
        ctime = (end-start)*100.
        start = time.time()
        for i in range(10):
            npks = localmaxlabel(im, l2 , wk, 1 )
        end=time.time()
        npks = localmaxlabel(im, l2 , wk, 11 )
        stime = (end-start)*100.
        start=time.time()
        for i in range(10):
            npks = localmaxlabel(im, l3 , wk, 2 )
        end=time.time()
        npks = localmaxlabel(im, l3 , wk, 12 )
        atime = (end-start)*100.
        if not (l2 == l3).all():
            import pylab as pl
            pl.imshow(l2)
            pl.imshow(l3)
            pl.show()
        self.assertEqual( (l2==l3).all(), True )  
        self.assertEqual( (l1==l3).all(), True )
        print("Timing 2048 c %.3f ms, sse %.3f, avx %.3f"%(
            ctime,stime,atime))

    def testbug(self):
        ''' case caught with image from Mariana Mar Lucas,
        caused a segfault'''
        for cpu in [0,1,2]:
            im = np.array([
                [ 0, 0, 0, 0, 0, 0, 0, 0],
                [ 0, 2, 0, 0, 0, 0, 0, 0],
                [ 0, 1, 1, 0, 0, 0, 0, 0],
                [ 0, 0, 0, 0, 0, 0, 0, 0],
                [ 0, 0, 0, 0, 0, 0, 0, 0],
                [ 0, 0, 0, 0, 0, 0, 0, 0]], np.float32)
            la = np.empty( im.shape, 'i' )
            wk = np.empty( im.shape, 'b' )
            localmaxlabel( im, la, wk, cpu )
            self.assertEqual(la[1,1],1)

    def testsmall(self):
        ''' sse / avx issue for tiny? '''
#        print("testsmall, fixme, segfault/infinite loop so no CI")
        for cpu in [0,1,2]:#11,12]:
            im = np.array([
                [ 0, 0, 0, 0, 0],
                [ 0, 2, 0, 0, 0],
                [ 0, 1, 1, 0, 0],
                [ 0, 0, 0, 0, 0]], np.float32)
            la = np.empty( im.shape, 'i' )
            wk = np.empty( im.shape, 'b' )
            localmaxlabel( im, la, wk, cpu )
            self.assertEqual(la[1,1],1)

        
if __name__== "__main__":
    import sys
    if len(sys.argv)>1:
        demo()
    unittest.main()

else:
    im = make_test_image(256)
    l2 = np.zeros( im.shape, 'i' )
    wk = np.zeros( im.shape, np.uint8 )
    npks = localmaxlabel(im, l2 , wk )
    





from __future__ import print_function

from ImageD11 import peakmerge, indexing, transformer
from ImageD11 import grain, unitcell, refinegrains
import sys, os, numpy as np, time, random

mytransformer = transformer.transformer()
#
# Your work starts here:
#    
mytransformer.loadfiltered( sys.argv[1] )
mytransformer.loadfileparameters(  sys.argv[2] )
tmp = sys.argv[3]
DSTOL = 0.003
OMEGAFLOAT = 0.13
NPKS = 80
TOLSEQ = [ 0.04, 0.01]
SYMMETRY = "cubic"
RING1 = [1,0]
RING2 = [1,]
# grid to search
translations = [(t_x, t_y, t_z) 
        for t_x in range(-500, 501, 200)
        for t_y in range(-500, 501, 200) 
        for t_z in range(-500, 501, 200) ]
# Cylinder: 
# filter( lambda x,y,z : (x*x+y*y)< 500*500, translations )
random.seed(42) # reproducible
random.shuffle(translations)


# printing or quiet:
if 1:
    NUL = open("/dev/null","w")
else:
    NUL = sys.stdout
    
# Get the unit cell
UC = unitcell.unitcell_from_parameters( mytransformer.parameterobj )


col = mytransformer.colfile
if not "drlv2" in col.titles:
    col.addcolumn( np.ones(col.nrows, np.float),
                   "drlv2" )
if not "labels" in col.titles:
    col.addcolumn( np.ones(col.nrows, np.float)-2,
                   "labels" )
if not "sc" in col.titles:
    assert "xc" in col.titles
    col.addcolumn( col.xc.copy(), "sc")
if not "fc" in col.titles:
    assert "yc" in col.titles
    col.addcolumn( col.yc.copy(), "fc")


class tick:
    def __init__(self):
        self.last = time.time()
        self.start = self.last
    def tick(self, msg=None):
        if msg is not None:
            sys.stdout.write(msg)
        now = time.time()
        sys.stdout.write(" %.2f total %.2f /s\n"%(now-self.last, now-self.start))
        self.last=now

ticker = tick()


def domap(  OmFloat,  OmSlop,
            pars,
            colfile,
            grainsfile,
            tolseq = [ 0.03, 0.02, 0.01],
            symmetry = "triclinic"):
    """mapping function - does what makemap.py does"""
    global NPKS
    ss = sys.stdout # turns off printing
    for tol in tolseq:
        sys.stdout = NUL
        o = refinegrains.refinegrains( OmFloat = OmFloat, OmSlop = OmSlop,
                                       tolerance = tol,
                                       intensity_tth_range = (0,180),
                                       )
        o.parameterobj = pars
        # o.loadfiltered ...
        o.scannames = ["internal"]
        o.scantitles = colfile.titles
        o.scandata["internal"] = colfile
        o.tolerance = tol
        o.readubis( grainsfile )
        if symmetry is not "triclinic":
            o.makeuniq( symmetry )
        o.generate_grains()
        o.refinepositions()
        gl = [x for x in list(o.grains.values()) if x.npks > NPKS]
        sys.stdout = ss
        if len(gl) == 0:
            print("I killed all your grains!")
            break
        else:
            print("Keeping",len(gl),"from",len(list(o.grains.values())),"grains with at least",NPKS,"peaks",tol)
            grain.write_grain_file( grainsfile ,  gl )
    return len(gl)



def doindex( gve, x, y, z, w):
    global NUL
    global tmp
    global NPKS
    global UC
    global TOLSEQ
    ss = sys.stdout # turns of printing
    sys.stdout = NUL
    myindexer = indexing.indexer(
        wavelength = w,
        unitcell = UC,
        gv = gve.T
        )
    myindexer.ds = np.sqrt( (gve * gve).sum(axis=0) )
    myindexer.ga = np.zeros(len(myindexer.ds),np.int)-1 # Grain assignments
    #   myindexer.readgvfile( gve )

    for ring1 in RING1:
        for ring2 in RING2:
            myindexer.parameterobj.set_parameters(  {
                'ds_tol': DSTOL, 
                'minpks': NPKS, 
                'max_grains': 1000, 
                'hkl_tol': TOLSEQ[0], 
                'ring_1': ring1,
                'ring_2': ring2
                } )
            myindexer.loadpars( )
            myindexer.assigntorings( )
            myindexer.find( )
            myindexer.scorethem( )
    grains = [grain.grain(ubi, [x,y,z]) for ubi in myindexer.ubis]
    grain.write_grain_file("%s.ubi"%(tmp),grains)
    sys.stdout = ss
    return len(grains)

    
### Computation starts here


os.system("cp %s %s.flt"%(sys.argv[1],tmp))

first = True

w =  mytransformer.parameterobj.get("wavelength")

totalgrains = 0

for t_x, t_y, t_z in translations:           
    mytransformer.updateparameters( )
    mytransformer.parameterobj.set_parameters(  {
                  't_x':t_x, 't_y':t_y, 't_z':t_z
                  } )
    mytransformer.compute_tth_eta( )
    if first:
        mytransformer.addcellpeaks( )
    mytransformer.computegv( )
    #    mytransformer.savegv( tmp+".gve" )
    gve = np.vstack(( mytransformer.colfile.gx,mytransformer.colfile.gy,mytransformer.colfile.gz ))
    if first:
        first=False
    ng = doindex(gve, t_x,t_y,t_z, w)
    print("Position",t_x, t_y, t_z,"Grains", ng, end=' ')
    ticker.tick()
    if ng > 0:
        #ret = os.system("makemap.py -u %s.ubi -U %s.map -s cubic --omega_slop=0.13 "%(tmp,tmp) +
        #                    "-f %s.flt  -t 0.02 -p %s  "%(
        #                       tmp,sys.argv[2]))
        nfit = domap( True,  OMEGAFLOAT,
                      mytransformer.parameterobj ,
                      mytransformer.colfile ,
                      "%s.ubi"%(tmp),
                      tolseq = TOLSEQ,
                      symmetry = SYMMETRY)
        if nfit > 0:
            open("all%s.map"%(tmp),"a").write(open("%s.ubi"%(tmp)).read())
            mytransformer.colfile.filter( mytransformer.colfile.labels < -0.1 )
            totalgrains += nfit
        ticker.tick()
    if  mytransformer.colfile.nrows == 0:
        print("All peaks indexed")
        break
    else:
        print("still got",mytransformer.colfile.nrows," to index, found",totalgrains,"in total")
            
        





#

import unittest
from ImageD11 import indexing
import sys

class fake():
    def write(self,*a,**k):
        pass
    def flush(self,*a,**k):
        pass

class testtwin( unittest.TestCase ):
    def testtwin(self):
        ss = sys.stdout
        sys.stdout = fake()
        myindexer = indexing.indexer()
        myindexer.readgvfile(  'Al.gve' )
        myindexer.updateparameters( )
        myindexer.parameterobj.set_parameters(
	{'minpks': 3, 'ring_1': 20, 'ring_2': 20} )
        myindexer.loadpars( )
        myindexer.assigntorings( )
        myindexer.find( )
        myindexer.scorethem( )
        sys.stdout = ss
        self.assertEqual( len(myindexer.ubis), 1)

if __name__=="__main__":
    unittest.main()




from __future__ import print_function, division
import sys
import numpy as np, fabio, time, os
from ImageD11 import cImageD11
print(cImageD11.__file__)

if 1:    
    M = 2048   # Overall image size
    N = 64     # tile size
    SIG = 3.   # sigma for peaks (width)
    # tile co-ordinates
    i,j = np.mgrid[0:N,0:N]-N/2.

    def pk(x,y,s):
        # 2D peak shape (with tails)
        dx = i-x
        dy = j-y
        gg= np.exp( -(dx*dx+dy*dy)/s/s )
        ll=s*s/(s*s+dx*dx+dy*dy)
        return ll+gg

    # Image for testing
    im = np.zeros( (M,M) )

    # Make a grid of points (tiles)
    pks = []
    for ii in range(0,M,N):
        for jj in range(0,M,N):
            # offset of centroids
            oi = 3*SIG*ii/M
            oj = SIG*ii/M
            # Ratio of heights
            sc = jj*1.0/M
            # put in two peaks, offset increases one direction,
            # scale in the other
            im[ii:ii+N,jj:jj+N] = pk( -oi, -oj, SIG ) + pk( oi, oj, SIG)*sc
            pks.append( (ii+N//2-oi,jj+N//2-oj) )
            pks.append( (ii+N//2+oi,jj+N//2+oj) )
    if 0:
        import pylab as pl
        pl.imshow( pl.log(im ))
        pl.show()

    s = 30000/im.max()
    int_im = (im*s).astype( np.uint16 )
    fabio.edfimage.edfimage( int_im ).write( "testoverlaps0000.edf" )

im = int_im
float_im=int_im.astype(np.float32)
labelc = np.zeros( im.shape, np.int32 )
labelm = np.zeros( im.shape, np.int32 )
work   = np.zeros( im.shape, np.int8 )


end0 = time.time()
nc    = cImageD11.connectedpixels( float_im, labelc, 1000 )
end1   = time.time()
rc    = cImageD11.blobproperties( float_im, labelc, nc, 0 )
cImageD11.blob_moments(rc)
end2   = time.time()
print(nc,"%.3f %.3f"%((end1-end0)*1000,(end2-end1)*1000))
wfloatim = np.where( float_im > 1000, float_im, float_im)
end0 = time.time()
nw    = cImageD11.localmaxlabel( wfloatim, labelm, work )
end1   = time.time()
rw    = cImageD11.blobproperties( wfloatim, labelm, nw, 0 )
cImageD11.blob_moments(rw)
end2   = time.time()
print(nw,"%.3f %.3f"%((end1-end0)*1000,(end2-end1)*1000))
if 0:
    pks=np.array(pks).T
    import pylab as pl
    pl.imshow( pl.log(float_im), origin='lower',
               interpolation='nearest', aspect='auto') 
    pl.plot( pks[1], pks[0],"+",label="ideal" )
    pl.plot( rc[:,cImageD11.f_raw],rc[:,cImageD11.s_raw],"x",
    label="connect")
    pl.plot( rw[:,cImageD11.f_raw],rw[:,cImageD11.s_raw],"o",
             label="watershed",
             markerfacecolor='None')
    pl.legend()
    pl.show()



import unittest

import numpy as np

from ImageD11 import labelimage
from ImageD11 import columnfile


class test_moments(unittest.TestCase):
    def setUp(self):
        self.data = np.array(
           [[20, 20,  0,  0,  2,  0,  0,  0],
            [20, 20,  0,  0,  0,  0,  6,  0],
            [ 0,  0,  0,  0,  0,  0,  0,  0],
            [ 0,  0,  0,  0,  4,  0,  0, 10],
            [ 1, 20,  0,  2,  4,  10,  0, 0],
            [ 1,  0,  0,  0,  5,  0,  0,  0],
            [ 1,  0,  0,  0,  10, 0,  0,  0],
            [ 1,  0,  0,  0,  0,  0,  0,  0]],
           np.float32
            )
        self.outfile = "momentstest.out"
        lio = labelimage.labelimage(self.data.shape , self.outfile)
        # print "go"
        lio.peaksearch(self.data, 0.1, 1.)
        lio.mergelast()
        lio.peaksearch(self.data, 0.1, 2.)
        lio.mergelast()
        lio.finalise()
        self.lio = lio


    def tearDown(self):
        try:
            os.remove("momentest.out")
        except:
            pass


    def testpyvsc(self):
        from ImageD11.cImageD11 import s_1, s_I, s_I2, \
            s_fI, s_ffI, s_sI, s_ssI, s_sfI, s_oI, s_ooI, s_foI, s_soI, \
            bb_mn_f, bb_mn_s, bb_mx_f, bb_mx_s, bb_mn_o, bb_mx_o, \
            mx_I, mx_I_f, mx_I_s, mx_I_o, \
            avg_i, f_raw, s_raw, o_raw, f_cen, s_cen, \
            m_ss, m_ff, m_oo, m_sf, m_so, m_fo
        from math import sqrt
        #print lio.res.shape, lio.lastres.shape
        #print "res",lio.res[:,m_sf]
        #print "lastres",lio.lastres[:,m_sf]
        for i in self.lio.res:
            #print i
            #import sys
            #sys.exit()
            if i[s_1] < 0.1:
                # Merged with another
                continue
            avg_i_p = i[s_I]/i[s_1]
            self.assertAlmostEqual( avg_i_p, i[avg_i] , 6 )
            # first moments
            fraw_p = i[s_fI]/i[s_I]
            self.assertAlmostEqual( fraw_p, i[f_raw] , 6 )

            sraw_p = i[s_sI]/i[s_I]
            self.assertAlmostEqual( sraw_p, i[s_raw] , 6 )

            oraw_p = i[s_oI]/i[s_I]
            self.assertAlmostEqual( oraw_p, i[o_raw] , 6 )

            # second moments - the plus 1 is the zero width = 1 pixel
            mss_p = sqrt( i[s_ssI]/i[s_I] - sraw_p*sraw_p + 1 )
            self.assertAlmostEqual( mss_p, i[m_ss] , 6 )

            mff_p = sqrt( i[s_ffI]/i[s_I] - fraw_p*fraw_p + 1 )
            self.assertAlmostEqual( mff_p, i[m_ff] , 6 )

            moo_p = sqrt( i[s_ooI]/i[s_I] - oraw_p*oraw_p + 1 )
            self.assertAlmostEqual( moo_p, i[m_oo] , 6 )

            msf_p = ( i[s_sfI]/i[s_I] - sraw_p*fraw_p )/mss_p/mff_p
            self.assertAlmostEqual( msf_p, i[m_sf] , 6 )

            mso_p = ( i[s_soI]/i[s_I] - sraw_p*oraw_p )/mss_p/moo_p
            self.assertAlmostEqual( mso_p, i[m_so] , 6 )

            mfo_p = ( i[s_foI]/i[s_I] - fraw_p*oraw_p )/mff_p/moo_p
            self.assertAlmostEqual( mfo_p, i[m_fo] , 6 )


class test_more_moments(test_moments):
    def setUp(self):
        self.data1 = np.array(
           [[20, 20,  0,  0,  2,  0,  0,  0],
            [20, 20,  0,  0,  0,  0,  6,  0],
            [ 0,  0,  0,  0,  0,  0,  0,  0],
            [ 0,  0,  0,  0,  4,  0,  0, 10],
            [ 1, 20,  0,  2,  4,  10,  0, 0],
            [ 1,  0,  0,  0,  5,  0,  0,  0],
            [ 1,  0,  0,  0,  10, 0,  0,  0],
            [ 1,  0,  0,  0,  0,  0,  0,  0]],
           np.float32
            )
        self.data2 = np.array(
           [[10, 00,  0,  0,  2,  0,  0,  0],
            [00, 20,  0,  0,  0,  0,  6,  0],
            [ 0,  0,  0,  0,  0,  0,  0,  0],
            [ 0,  0,  0,  0,  4,  0,  0, 10],
            [ 1, 20,  0,  2,  4,  10,  0, 0],
            [ 1,  0,  0,  0,  5,  0,  0,  0],
            [ 1,  0,  0,  0,  10, 0,  0,  0],
            [ 1,  0,  0,  0,  0,  0,  0,  0]],
           np.float32
            )
        self.data3 = np.array(
           [[00, 00,  0,  0,  2,  0,  0,  0],
            [00, 10,  0,  0,  0,  0,  6,  0],
            [ 0,  0,  0,  0,  0,  0,  0,  0],
            [ 0,  0,  0,  0,  4,  0,  0, 10],
            [ 1, 20,  0,  2,  4,  10,  0, 0],
            [ 1,  0,  0,  0,  5,  0,  0,  0],
            [ 1,  0,  0,  0,  10, 0,  0,  0],
            [ 1,  0,  0,  0,  0,  0,  0,  0]],
           np.float32
            )
        self.outfile = "momentstest.out"
        lio = labelimage.labelimage(self.data1.shape , self.outfile)
        # print "go"
        lio.peaksearch(self.data1, 0.1, 1.)
        lio.mergelast()
        lio.peaksearch(self.data2, 0.1, 2.)
        lio.mergelast()
        lio.peaksearch(self.data3, 0.1, 3.)
        lio.mergelast()
        lio.finalise()
        self.lio = lio


    def tearDown(self):
        try:
            os.remove("momentest.out")
        except:
            pass



class test_labelimage(unittest.TestCase):
    def setUp(self):
        self.dims = (200,300)
        self.data1 = np.zeros(self.dims, np.float32)
        self.data2 = np.zeros(self.dims, np.float32)
        self.data3 = np.zeros(self.dims, np.float32)
        self.data4 = np.zeros(self.dims, np.float32)
        self.data5 = np.zeros(self.dims, np.float32)
        self.outfile = "l2.out"


    def test_1(self):
        self.data1[20:41,120:141] = 1.
        self.data2[20:41,120:141] = 1.
        self.data3[20:41,120:141] = 1.
        self.data4[20:41,120:141] = 1.
        self.data5[20:41,120:141] = 1.
        self.data3[30,130] = 2. # in the middle
        self.data1[55,65] = 1.
        self.data2[55,65] = 2.
        self.data3[55,65] = 3.
        self.data4[55,65] = 2.
        self.data5[55,65] = 1.

        lio = labelimage.labelimage(self.dims , self.outfile)
        lio.peaksearch(self.data1, 0.1, 1.)
        lio.mergelast()
        lio.peaksearch(self.data2, 0.1, 2.)
        lio.mergelast()
        lio.peaksearch(self.data3, 0.1, 3.)
        lio.mergelast()
        lio.peaksearch(self.data4, 0.1, 4.)
        lio.mergelast()
        lio.peaksearch(self.data5, 0.1, 5.)
        lio.mergelast()
        lio.finalise()
        lio.outfile.close()

        co = columnfile.columnfile(self.outfile)



        self.assertEqual(co.nrows,2)
        # first peaks 20->41, 120->141
        self.assertAlmostEqual(co.fc[0],130.,6)
        self.assertAlmostEqual(co.sc[0],30.,6)
        self.assertAlmostEqual(co.omega[0],3.0,6)
        self.assertAlmostEqual(co.Number_of_pixels[0],21*21*5,6)
        self.assertAlmostEqual(co.avg_intensity[0],1,1)
        self.assertAlmostEqual(co.IMax_int[0], 2 , 6)
        self.assertAlmostEqual(co.IMax_f[0], 130., 6)
        self.assertAlmostEqual(co.IMax_s[0], 30.0, 6)
        self.assertAlmostEqual(co.IMax_o[0], 3.0, 6)

        self.assertAlmostEqual(co.fc[1],65.,6)
        self.assertAlmostEqual(co.sc[1],55.,6)
        self.assertAlmostEqual(co.omega[1],3.0,6)
        self.assertAlmostEqual(co.Number_of_pixels[1],5,6)
        self.assertAlmostEqual(co.avg_intensity[1],1.8,1)
        self.assertAlmostEqual(co.IMax_int[1], 3 , 6)
        self.assertAlmostEqual(co.IMax_f[1], 65., 6)
        self.assertAlmostEqual(co.IMax_s[1], 55., 6)
        self.assertAlmostEqual(co.IMax_o[1], 3.0, 6)


        mask = co.Number_of_pixels > 10
        co.filter(mask)
        self.assertEqual(co.nrows, 1)
        co.writefile("l1.out")


if __name__=="__main__":
    unittest.main()



#!/usr/bin/env python

from __future__ import print_function


# ImageD11_v1 Software for beamline ID11
# Copyright (C) 2008  Jon Wright
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  0211-1307  USA

import os, sys, logging
from argparse import ArgumentParser
from ImageD11 import ImageD11options
from ImageD11 import fft_index_refac, rc_array,\
    lattice_reduction, indexing
import numpy as np
from ImageD11.fft_index_refac import grid


def get_options(parser):
    parser.add_argument('-g', '--gve',
                      action = 'store',
                      type=ImageD11options.GvectorFileType(mode='r'),
                      dest = 'gvefilename',
                      default = None,
                      help = "Filename for g-vectors")
    parser.add_argument('-k', '--ngrains',
                      action = 'store',
                      dest = 'ngrains',
                      type = int,
                      default = 1,
                      help = "number of grains to try to find")
    parser.add_argument('-o', '--output',
                      action = 'store',
                      type=ImageD11options.UbiFileType(mode='w'),
                      default = 'grains.ubi',
                      dest = 'outfile',
                      help = "Name of ubi file to save grains in")

    parser = lattice_reduction.get_options(parser)

    parser.add_argument('--fft',
                      action = 'store_true',
                      dest = 'use_fft',
                      default = False,
                      help = "Use fft to generate lattice vectors [False]")

    parser.add_argument('--score_fft',
                      action = 'store_true',
                      dest = 'score_fft',
                      default = False,
                      help = "Score fft peaks using fft peaks first [True]")

    parser.add_argument('--no_sort',
                      action = 'store_false',
                      dest = 'sort_gve',
                      default = True,
                      help = "Sorting the gvector by length before indexing [True]")

    parser.add_argument('--noisy',
                      action = 'store_true',
                      dest = 'noisy',
                      default = False,
                      help = "Print more output")
    fft_index_refac.get_options( parser )
    return parser






if __name__=="__main__":

    logging.basicConfig( level=logging.INFO )
    parser = get_options( ArgumentParser() )

    options = parser.parse_args()
    
    
    o = indexing.indexer()
    if options.gvefilename is None or \
            not os.path.exists(options.gvefilename):
        print("You need to supply a gvector file with the -g option")
        sys.exit()

    o.readgvfile(options.gvefilename)


    mags = [ [np.dot( g, g ), tuple(g)] for g in o.gv]
    mags.sort()
    sorted = np.array( [ np.array(g[1]) for g in mags ] )
    
    all_gvecs = rc_array.rc_array(sorted.copy(), direction = 'row' )
    cur_gvecs = rc_array.rc_array(sorted.copy(), direction = 'row' )

    ubis = []
    
    try:
        for i in range(options.ngrains):
            if len(cur_gvecs) < 3:
                print("Ran out of unindexed peaks")
                break

            # print "Peak remaining",len(cur_gvecs)
            if options.use_fft:
                # do fft
                g = grid( npx = options.npx,
                          mr = options.mr,
                          nsig = options.nsig)
                g.gv_to_grid_new(cur_gvecs)
                g.fft()
                g.props()
                g.peaksearch(open("eu.patterson_pks","w"))
                g.read_peaks("eu.patterson_pks")
                vecs = rc_array.rc_array(g.UBIALL.T , direction='col')
                assert vecs.shape == (3, len(g.UBIALL))
                order = np.argsort( g.colfile.sum_intensity )[::-1]
                vecs = rc_array.rc_array( np.take( vecs, order, axis = 1),
                                          direction = 'col')
                assert g.gv.shape[1] == 3
                print("Finding lattice from patterson")
                # Go through and make a shortlist of lattices by scoring fft only
                if options.score_fft:
                    test_set = vecs
                else:
                    test_set = all_gvecs
                    #for v in vecs[:options.n_try]:
                    #     print v,np.sqrt(np.dot(v,v))
                try:
                    # mv2 = 1/options.min_vec2
                    l = lattice_reduction.find_lattice(
                        vecs,
                        min_vec2 = options.min_vec2,
                        n_try = options.n_try,
                        test_vecs = test_set,
                        tol = options.tol,
                        fraction_indexed = options.fraction_indexed,
                        noisy = options.noisy,
                        )
                except IndexError:
                    print(vecs, options.n_try)
                    print(vecs.shape)
                    raise
            else:
                # Test vectors are the g-vectors
                # Sort by length before searching??
                logging.info("Doing lattice search")
                l = lattice_reduction.find_lattice(
                    cur_gvecs, 
                    min_vec2=options.min_vec2,
                    n_try=options.n_try,
                    test_vecs = all_gvecs, 
                    tol = options.tol,
                    fraction_indexed = options.fraction_indexed,
                    noisy = options.noisy,
                    )
            if l is None:
                break
            l.r2c = indexing.refine( l.r2c, all_gvecs, tol = options.tol)
            l.c2r = np.linalg.inv(l.r2c)

            #print "all_gv 3,4", all_gvecs[3:5]
            #print np.dot(l.r2c, all_gvecs[3:5].T)
            #print l.score(all_gvecs[3:5], tol = options.tol)
            #print l.remainders( all_gvecs[3:5])
                        
            all = len(all_gvecs)
            npks = l.score( all_gvecs , tol = options.tol )
            if (1.0*npks/len(all_gvecs)) < options.fraction_indexed:
                break
            dr = indexing.calc_drlv2( l.r2c, all_gvecs) 
            t2 = options.tol * options.tol
            # print "tol, dr",t2,dr.shape, [d for d in dr[:10]] # print it !!!
            n2 = np.sum(np.where(dr < t2 , 1, 0 ))
            n3 = np.sum(np.where(dr > t2 , 1, 0 ) )
            assert npks == n2 , "debug scoring"
            assert n3 == len(all_gvecs) - npks, "debug scoring"
            print(l.r2c)
            print("Unit cell:",(6*"%.6f ")%indexing.ubitocellpars(l.r2c))
            
            # Put this grain in the output list
            ubis.append(l.r2c)

            # Remove from the gvectors
            drlv2 = indexing.calc_drlv2( l.r2c, cur_gvecs )
            # print drlv2[:20]
            # print cur_gvecs.shape
            # print drlv2.shape,drlv2[:10],options.tol*options.tol
            
            cur_gvecs = rc_array.rc_array(
                np.compress( drlv2 > options.tol*options.tol,
                             cur_gvecs, axis = 0),
                direction = 'row')
            print("Lattice found, indexes", npks, "from all",all)
            print("Number of unindexed peaks remaining %d"%(len(cur_gvecs)))
            print("Current vector shape",cur_gvecs.shape)
        if len(ubis)>0:
            indexing.write_ubi_file(options.outfile,ubis)
            print("Wrote to file",options.outfile)
        else:
            print("No unit cell found, sorry, please try again")
    except:
        if len(ubis)>0:
            indexing.write_ubi_file(options.outfile,ubis)
            print("Wrote to file",options.outfile)
        raise 



#!/usr/bin/env python

from __future__ import print_function

# 
# ImageD11_v1.0 Software for beamline ID11
# Copyright (C) 2005-2009  Jon Wright
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  0211-1307  USA



from six.moves import input


"""
Script for repairing use of incorrect spline file
"""
from ImageD11 import columnfile, blobcorrector


def fix_flt( inname, outname , cor ):
    inc = columnfile.columnfile( inname )
    for i in range( inc.nrows ):
        inc.sc[i], inc.fc[i] = cor.correct( inc.s_raw[i], inc.f_raw[i] )
    inc.writefile( outname )

def help():
    print(sys.argv[0] + " columnfile_in columnfile_out splinefile")
    print("use splinefile name perfect to remove distortion")
    print("replaces fc and sc columns with corrected numbers")
        

def mymain():
    # If we are running from a command line:
    import sys, os
    inname = sys.argv[1]
    if not os.path.exists(inname) or len(sys.argv) < 4:
        help()
        sys.exit()
    outname = sys.argv[2]
    if os.path.exists(outname):
        if not input("Sure you want to overwrite %s ?"%(outname)
                         )[0] in ['y','Y']:
            sys.exit()
    splinename = sys.argv[3]
    if splinename == 'perfect':
        cor = blobcorrector.perfect()
    else:
        cor = blobcorrector.correctorclass( splinename )

    fix_flt( inname, outname, cor )

if __name__=="__main__":
    mymain()




from __future__ import print_function

"""
Module for read saint raw files from bruker integration
"""
import numpy
from ImageD11 import columnfile


docsheader = """
...from the file sairefl._tp...
Copyright Bruker

ITEM                  FORMAT   DESCRIPTION
----                  ------   -----------

COMMENT                 '!'    An exclamation mark in the first column
                               indicates that the line contains a
                               comment.  (If the line is not a
                               comment, the record starts in the first
                               column, not the second; that is, the
                               first column is not "reserved" for a
                               possible exclamation mark)
"""

docs = """
IHKL                   3I4     HKL indices of the reflection      

  #IMNP                   3I4     PRESENT ONLY FOR MODULATED-STRUCTURE DATA
                               MNP indices of the reflection.  Unused
                               indices  (in 4- or 5-dimensional cases)
                               are written as zero

FI                     F8.x    Integrated intensity in photons, background-
                               subtracted, normalized to 1 min/deg, and
                               Lp corrected.  Number of places to right
                               of the decimal point is adjusted to
                               balance precision vs. overflow;  exponential
                               format is used for very large values.

SI                     F8.x    Estimated standard deviation of the
                               intensity.  Number of places to right
                               of the decimal point is adjusted to
                               balance precision vs. overflow;  exponential
                               format is used for very large values.

IBATNO                 I4      Batch number (for scaling) assigned to
                               the reflection

COSINES                6F8.5   Direction cosines of the incident and
                               diffracted beams, relative to the 
                               unit cell axes.  Order is XI,XD,YI,YD,
                               ZI,ZD where XI is the X-component of
                               the incident beam, XD is the X-component
                               of the diffracted beam, etc.  Used
                               for absorption correction.


MSTATUS                I3      1X,Z2   Status mask reserved for flagging
                               abnormal conditions.  There are
                               currently none defined.

XO                     F7.2    Observed X-pixel coordinate of the
                               intensity-weighted reflection centroid,
                               in reduced pixels (scale of 512x512)

YO                     F7.2    Observed Y-pixel coordinate of the
                               intensity-weighted reflection centroid,
                               in reduced pixels (scale of 512x512)

ZO                     F8.2    Observed frame number of the
                               intensity-weighted reflection centroid

XP                     F7.2    Predicted X-pixel of the reflection
                               in reduced pixels (scale of 512x512)

YP                     F7.2    Predicted Y-pixel of the reflection
                               in reduced pixels (scale of 512x512)

ZP                     F8.2    Predicted frame number of the reflection

CORLPAF                F6.3    Multiplicative correction for Lorentz
                               effect, polarization, air absorption,
                               and detector faceplate absorption,
                               already applied to integrated intensity,
                               FI

CORREL                 F5.2    The correlation coefficient between the
                               3-D profile observed for this reflection
                               and the corresponding model 3-D profile

ACCUMTIME              F7.2    Accumulated hours of exposure

SWING                  F7.2    Detector swing angle in degrees

ROT                    F7.2    Scan axis (omega or phi) setting in
                               degrees at which the reflection was
                               observed

IAXIS                  I2      Scan axis number (2=omega, 3=phi)

ISTL                   I5      SIN(THETA)/LAMBDA times 10,000 

PKSUM                  I9      Total raw peak counts in photons, with no
                               correction for background, normalization, 
                               or Lp

BGAVG                  I7      Normally, average BG per pixel in 
                               photons * 1000. In data sets where the 
                               background was reported to be very large 
                               ( > 1024 photons after 1 min/deg 
                               normalization), the program issues a warning 
                               message during integration and the scale of
                               1000 is omitted.

ROTREL                 F7.2    Rotation of the scan axis (omega or phi)
                               in degrees relative to the start of the
                               run

ICRYST                 I4      Crystal number (for scaling)

PKFRAC                 F6.3    Fraction of the profile volume nearest
                               this HKL.  1-PKFRAC is the fraction of
                               the intensity which had to be estimated
                               from the model profiles because of
                               overlap with neighboring spots

IKEY                   I11     The unique sort key for the group of
                               equivalents to which this HKL belongs.
                               IKEY = 1048576 * (H+511) + 1024 *
                               (K+511) + (L+511), where H,K and L are
                               the HKL indices transformed to the base
                               asymmetric unit.

IMULT                  I3      The point-group multiplicity of this HKL

CORLORENTZ             F6.3    Lorentz correction

XGEO                   F8.2    Spatially corrected X relative to beam
                               center, in pixels

YGEO                   F8.2    Spatially corrected Y relative to beam 
                               center, in pixels

CHI                    F8.3    Chi setting angle, degrees

OTHER_ANGLE            F8.3    The remaining setting angle, degrees.  This
                               will be phi if scans were in omega, or omega
                               if scans were in phi

ICOMPONENT             I4      Twin component number in SHELXTL HKLF 5
                               convention.  In a multi-component overlap,
                               ICOMPONENT is negated for all but the last
                               record of the overlap

"""

    
    

class saintraw(object):
    doc = docs
    titles  = []
    formats = {}
    helps   = {}
    def __init__(self, filename=None):
        """
        filename = filename to read in
        """
        self.parsedocs()
        if filename is not None:
            self.read(filename)
    
    def parsedocs(self):
        """
        Parse the saint documentation for the Bruker format
        """
        self.titles = []
        title = None
        for line in self.doc.split("\n"):
            if len(line.rstrip()) == 0:
                if title is not None:
                    self.formats[title] = format
                    self.helps[title]   = help
                    self.titles.append(title)
                    title = None
                    format = None
                continue
            if line[0] != " ":
                title, format = line.split()[0:2]
                help = " ".join(line.split()[2:])
            else:
                help = " ".join([help, line.lstrip()])

        alltitles = []
        slices = []
        funcs = []
        i = 0
        allformats = []
        for t in self.titles:
            f = self.formats[t]
            if f[0].isdigit():
                n = int(f[0])
                f = f[1:]
            else:
                n = 1
            if n > 1:
                for j in range(n):
                    alltitles.append( t + "_%d" % (j) )
                    allformats.append( f )
            else:
                alltitles.append( t )
                allformats.append( f )
            assert f[0] in ["I","F"]
            if f[0] == "I":
                for dummy in range(n):
                    funcs.append( int )
            if f[0] == "F":
                for dummy in range(n):
                    funcs.append( float )
            num = int(f[1:].split(".")[0])
            for dummy in range(n):
                slices.append( slice( i, i + num ) )
                i += num
        self.alltitles = alltitles
        self.allformats = allformats
        self.funcs = funcs
        self.slices = slices
        assert len(funcs) == len(slices)
        assert len(slices) == len(alltitles)
            
    def read(self, filename):
        """
        Read an ascii formatted saint reflection file
        """
        self.data = {}
        self.lines = open(filename,"r").readlines()
        for t in self.alltitles:
            self.data[t] = []
        zipped = list(zip(self.alltitles, self.slices, self.funcs))
        for line in self.lines:
            if line[0] == "!":
                # Comment line
                continue
            for t,s,f in zipped:
                # Parse this line
                try:
                    self.data[t].append( f( line[s] ) )
                except:
                    print(t,s,f)
                    raise

    def condition_filter(self, name, func):
        """
        Remove the peaks according to condition
        """
        assert len(self.lines) == len(self.data[name] )
        indices = numpy.compress( func( numpy.array( self.data[name]) ) , 
                                  list(range(len(self.lines))) )
        self.take( indices )                                

    def take(self, order):
        """
        Put the peaks in the order given in order (indices)
        """
        for t in list(self.data.keys()):
            self.data[t] = numpy.take( self.data[t],
                                       order) 
        self.lines = list( numpy.take( self.lines,
                                       order))
    
    def sort(self, name):
        """
        Sort according to a column in self.data
        """
        order = numpy.argsort( self.data[name] )
        self.take(order) 


    def write(self, filename):
        """
        Write an ascii formatted saint reflection file

        """
        outf = open(filename, "w")
        for line in self.lines:
            outf.write( line )
        # raise Exception("Not implemented writing yet!")

    def tocolumnfile(self):
        """
        Return a columnfile
        """
        cof = columnfile.newcolumnfile( self.alltitles )
        dlist = [ self.data[t] for t in self.alltitles ]
        cof.bigarray = numpy.array( dlist, numpy.float )
        cof.nrows = len( self.data[ self.alltitles[0] ] )
        cof.ncols = len( self.alltitles ) 
        cof.set_attributes()
        return cof

if __name__ == "__main__":

    import sys, time
    START = time.time()
    sra = saintraw()
    print("Making object", time.time() - START)
    
    START = time.time()
    sra.read(sys.argv[1])
    print("Reading", time.time() - START)
    
    print(len(sra.data['IHKL_0']))

    START = time.time()
    cra = sra.tocolumnfile()
    print(cra.bigarray.shape)
    print("Convert to colfile", time.time() - START)



#!/usr/bin/env python

from __future__ import print_function

from ImageD11.grain import read_grain_file
import sys, os

gf = read_grain_file(sys.argv[1])
mapfile=open(sys.argv[2],"w")

def dodot(xyz,k):
    mapfile.write("%f %f %f %d\n"%(xyz[0],xyz[1],xyz[2],k))

def getmedian(s):
    items=s.split()
    j = -1
    for i in range(len(items)):
        if items[i] == "median":
            j = i
    if j == -1:
        return 0
    return abs(float(items[j+2]))
            
try:
    outersf = float(sys.argv[3])
except:
    outersf = 1.0

print("Scale factor is",outersf)
for g in gf:
    #print g.translation, g.ubi
    mapfile.write("\n\n")
    o = g.translation
    try:
        sf = pow(getmedian(g.intensity_info),0.3333)*outersf
    except:
        sf = outersf
    try:
        k = int(g.npks)
    except:
        k = 1
    for u in g.ubi:
        dodot(o,k)
        dodot(o+u*sf,int(g.npks))
    for u in g.ubi:
        dodot(o,k)
        dodot(o-u*sf,int(g.npks))
#    dodot(o,k)
#    dodot(o+sf*(-g.ubi[0]-g.ubi[1]),k)
#    dodot(o,k)
#    dodot(o+sf*(g.ubi[0]+g.ubi[1]),k)

mapfile.close()
term = " "
if "linux" in sys.platform:
    term = "set term x11"
if "win32" in sys.platform:
    term = "set term windows"
    
open("gnuplot.in","w").write("""
%s
set ticslevel 0
set title "Color proportional to number of peaks"
set palette model RGB
set palette defined ( 0 "violet", 1 "blue", 2 "green", 3 "yellow", 4 "orange", 5 "red" )
set view equal xyz
set view 75,0,1,1
#set terminal gif animate delay 10 loop 1 optimize size 1024,768
set nokey
set hidden3d
#set output "ImageD11map.gif"
splot "%s" u 1:2:3:4 w l lw 2 lc pal z
"""%(term, sys.argv[2])
# "".join(["set view 75,%d\n replot\n"%(i) for i in range(1,360,1)])
                             )


    
os.system("gnuplot -background white gnuplot.in -")

    





from __future__ import print_function


from ImageD11 import sym_u
from ImageD11 import grain
import sys, numpy as np, xfab.tools

print("# Usage: %s grains1.map grains2.map group angletol  distancetol"%(sys.argv[0]))
print("# Matches grains using symmetry group from:  ")
print("#  [cubic|hexagonal|trigonal|rhombohedralP|tetragonal|orthorhombic|monoclinic_[a|b|c]|triclinic]")


g1l = grain.read_grain_file(sys.argv[1])
g2l = grain.read_grain_file(sys.argv[2])
for g in g1l + g2l:
    if g.translation is None:
        g.translation = np.zeros(3)
try:
    h = getattr( sym_u, sys.argv[3] )()
except:
    print("# No group!")
    print("#Using cubic")
    h = sym_u.cubic()
print("# Symmetry considered")
for o in h.group:
    print(o)

try:
    tolangle = float(sys.argv[4])
except:
    tolangle = 1.0

try:
    toldist = float(sys.argv[5])
except:
    toldist = 100.0

dt0 = np.array([0.,0.,0.])
if len(sys.argv)>=9:
    try:
        dt0 = np.array([float(x) for x in sys.argv[6:9]])
    except:
        raise



da=np.zeros( (len(g1l),len(g2l)), np.float)
dt2=np.zeros( (len(g1l),len(g2l)), np.float)

for g in g1l + g2l:
    g.u = xfab.tools.ubi_to_u_b(g.ubi)[0]
    assert (abs(np.dot(g.u, g.u.T) - np.eye(3)).ravel()).sum() < 1e-6
    
dtsum = np.zeros(3)
ndt = 0
for i,g1 in enumerate(g1l):
    minangle = 180.0
    best = None
    symubis = [np.dot(o, g1.ubi) for o in h.group]
#    symus   = [xfab.tools.ubi_to_u_b(ubi)[0] for ubi in symubis]
    asymusT   = [xfab.tools.ubi_to_u_b(ubi)[0].T for ubi in symubis]
    printed = False
    trace = np.trace
    pi = np.pi
    for j,g2 in enumerate(g2l):
        sg = None
        aumis = np.dot(asymusT, g2.u)
        # print aumis.shape
        # trace(aumis,axis1=1,axis2=2)
        arg = (aumis[:,0,0]+aumis[:,1,1]+aumis[:,2,2] - 1. )/2.
        asymangle = np.arccos(np.clip(arg, -1, 1))
        sg = np.argmin(asymangle)
        angle = asymangle[sg]*180.0/pi
        # translation
        dt = g1.translation - g2.translation + dt0
        da[i,j] = angle
        dt2[i,j] = np.sqrt((dt*dt).sum())
        if angle < tolangle and dt2[i,j] < toldist:
            #print "# Found a match!", h.group[sg].ravel()
            dtsum += dt
            ndt += 1
            print(i,j,"angle  %.4f"%(angle),"distance %.3f"%(dt2[i,j]),"\t", end=' ')
            print(3*"%.1f "%tuple(dt),"\t", end=' ')
            print(3*"%.1f "%tuple(dtsum/ndt))
            #print "# grain ubi,t"
            #print g1.ubi,g1.translation
            #print g2.ubi,g2.translation
            printed = True
    if not printed:
        ja = np.argmin(da[i])
        jd = np.argmin(dt2[i])
        print("# not matched %d min angle %d  %.4f  %.3f"%(i,ja,da[i,ja],dt2[i,ja]))
        print("# not matched %d min dist  %d  %.4f  %.3f"%(i,ja,da[i,jd],dt2[i,jd]))


np.save( "da",da)
np.save( "dt2",dt2)




from __future__ import print_function
import math, numpy, time
from ImageD11 import cImageD11
from fabio.openimage import openimage
print("Using class version")




class fourier_radial(object):
    """ Cache results for re-use where possible on next layer """
    def __init__(self, dims, theta=None):
        self.dims = dims
        self.theta = theta
        if self.theta is not None:
            assert len(self.theta) == dims[1]
            self.theta = numpy.array(theta)*numpy.pi/180.0
            self.make_indices()

    def set_theta(self, theta):
        """
        th is the list of angles in degrees of the projections
        assumed 1 degree steps otherwise
        """
        self.theta = theta
        assert len(self.theta) == self.dims[1]
        self.theta = numpy.array(theta)*numpy.pi/180.0
        self.make_indices()

    def make_indices(self):
        arshape = self.dims[0]/2+1, self.dims[1]
        nv = (arshape[0]-1)*2
        nh = arshape[0]
        print("NV,NH",nv,nh)
        self.ftimshape = (nv, nh)
        self.ftimlen = nv*nh
        n1 = (self.dims[0]/2+1)*self.dims[1]
        xv   = numpy.arange(0, self.dims[0]/2+1, 1, 
                            dtype=numpy.float32 )
        # dimensions?
        cth = numpy.cos( self.theta ) # 1D
        sth = numpy.sin( self.theta ) # 1D
        ia = numpy.round(numpy.outer( cth, xv )).astype(numpy.int)
        ja = numpy.round(numpy.outer( sth, xv )).astype(numpy.int)
        on = numpy.array([1.0],numpy.float32)
        jm = numpy.where(ja < 0, -on, on)
        numpy.multiply( ia, jm, ia ) #  if j<0: i=-i
        numpy.multiply( ja, jm, ja ) #  if j<0: j=-j
        #  if j<0: f=f.conj()
        ia = numpy.where( ia < 0, nv+ia, ia)
        inds = (ia*nh + ja).ravel()
        self.conjer = jm
        self.inds = inds
        nim  = numpy.zeros( ( nv* nh), numpy.float32 )
        wons = numpy.ones( (len(inds)), dtype=numpy.float32 )
        # This is now more dense - bincount?
        cImageD11.put_incr( nim , inds, wons )
        nim = nim.astype(numpy.int)
        self.nim_div = nim + (nim==0)
        

    def process_sinogram( self,
                          sinogram, 
                          do_interpolation=False):

        """
        sinogram is from the data
            dimensions [npixels, nangles]
            do_interp - tries to fill in some of the missing data in
            fourier space
        returns the radon transform
        """
        assert sinogram.shape == self.dims
        ar = numpy.fft.rfft(sinogram, axis=0)
        faprojr = (ar.T.real.astype(numpy.float32))
        faprojc = (ar.T.imag.astype(numpy.float32))
        numpy.multiply( faprojc, self.conjer, faprojc)
        fimr = numpy.zeros( self.ftimlen , numpy.float32 )
        fimc = numpy.zeros( self.ftimlen , numpy.float32 )
        cImageD11.put_incr( fimr, self.inds, faprojr.ravel())
        cImageD11.put_incr( fimc, self.inds, faprojc.ravel())
        fim = fimr + fimc*1j
        fim = numpy.divide( fimr + fimc*1j, self.nim_div)
        fim.shape = self.ftimshape 
        return fim

    def sino2im(self, sinogram, centrepixel ):
        # Take out high frequency in mean (some ring artifacts)
        s = sinogram
        cp = centrepixel
        d  = numpy.concatenate( (  s[cp:,:], s[:cp,:], ), axis=0)
        im = self.process_sinogram( d , centrepixel )    
        # figure out what the scale factor really is
        ret = numpy.fft.irfft2( im ) * im.shape[0] * im.shape[1]
        ret = numpy.fft.fftshift( ret )
        return ret


if __name__=="__main__":
    import sys
    if len(sys.argv) != 5:
        print("Usage: sinogram startangle step centrepixel")
        sys.exit()
    
    fname = sys.argv[1]
    star = time.time()
    sino = openimage( fname )
    na,nx = sino.data.shape
    start = float(sys.argv[2])
    step  = float(sys.argv[3])
    
    centrepixel = int( sys.argv[4] )
    end   = na*step + start
    print("start, step, end",start, step, end)
    angles = numpy.arange(start, end, step)
    assert len(angles) == na,"%d %d  ... %d"%(nx,na,len(angles))
    print("%.2f setup"%(time.time()-star))
    d =  sino.data.T[:1200]
    o = fourier_radial( d.shape, angles )
    start = time.time()
    im = o.sino2im( d, centrepixel )
    sino.data = im
    sino.write(fname+"_r", force_type=numpy.float32)
    import pylab
    pylab.imshow(im, interpolation='nearest',aspect='auto')
    pylab.show()
    print("per image",time.time()-start)





from __future__ import print_function


from numpy import arange, ones, concatenate
from ImageD11.columnfile import columnfile
import glob, sys
n = 0
allc=None


for f in sys.argv[2:]:
        c = columnfile(f)
        c.filter(c.Number_of_pixels > 4)
        #print f.split("_")
        nums = int(f.split("_")[-3])
        print(f,nums)
        c.addcolumn(ones(c.nrows)*nums,"dty")
        if allc is None:
            allc = c.copy()
        else:
            allc.bigarray = concatenate( (allc.bigarray, c.bigarray), axis=1 )
            allc.nrows = allc.bigarray.shape[1]
            allc.set_attributes()
        
allc.writefile(sys.argv[1])



#!/usr/bin/env python
from __future__ import print_function


"""
Removes grains from a grain file if they have less than
npks peaks

Usage : cutgrains.py   ubi_in  ubi_out  npks
"""

from ImageD11.grain import read_grain_file, write_grain_file
import sys
try:
    GRAINS = read_grain_file(sys.argv[1])
    NPKS = int(sys.argv[3])
    KEEP = []
    for g in GRAINS:
        if int(g.npks) >= NPKS:
            KEEP.append(g)
    write_grain_file( sys.argv[2], KEEP)
except:
    print(__doc__)
    raise
            




from __future__ import print_function

#!/usr/bin/env python
#


""" Simplex - a regression method for arbitrary nonlinear function minimization

Simplex minimizes an arbitrary nonlinear function of N variables by the
Nelder-Mead Simplex method as described in:

Nelder, J.A. and Mead, R., "A Simplex Method for Function Minimization",
   Computer Journal, Vol. 7, 1965, pp. 308-313.

It makes no assumptions about the smoothness of the function being minimized.
It converges to a local minimum which may or may not be the global minimum
depending on the initial guess used as a starting point.



# Copyright (c) 2001 Vivake Gupta (v@omniscia.org).  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA
#
# This software is maintained by Vivake (v@omniscia.org) and is available at:
#     http://www.omniscia.org/~vivake/python/Simplex.py

# Modified (debugged?) 7/16/2004 Michele Vallisneri (vallis@vallis.org)

# Copied into ImageD11 by Jon Wright - This works really really well!!!
"""

import math
import copy
import sys

class Simplex:
    def __init__(self, testfunc, guess, increments, kR = -1, kE = 2, kC = 0.5):
        """Initializes the simplex.
        INPUTS
        ------
        testfunc      the function to minimize
        guess[]       an list containing initial guesses
        increments[]  an list containing increments, perturbation size
        kR            reflection constant  (alpha =-1.0)
        kE            expansion constant   (gamma = 2.0)
        kC            contraction constant (beta  = 0.5)
        """
        self.testfunc = testfunc
        self.guess = guess
        self.increments = increments
        self.kR = kR
        self.kE = kE
        self.kC = kC
        self.numvars = len(self.guess)
        self.simplex = []

        self.lowest = -1
        self.highest = -1
        self.secondhighest = -1

        self.errors = []
        self.currenterror = 0

        # Initialize vertices
        # MV: the first vertex is just the initial guess
        #     the other N vertices are the initial guess plus the individual increments
        #     the last two vertices will store the centroid and the reflected point
        #     the compute errors at the ... vertices

        for vertex in range(0, self.numvars + 3):
            self.simplex.append(copy.copy(self.guess))

        for vertex in range(0, self.numvars + 1):
            for x in range(0, self.numvars):
                if x == (vertex - 1):
                    self.simplex[vertex][x] = self.guess[x] + self.increments[x]
            self.errors.append(0)
        self.calculate_errors_at_vertices()

    def minimize(self, epsilon = 0.0001, maxiters = 250, monitor = 1):
        """Walks to the simplex down to a local minima.
        INPUTS
        ------
        epsilon       convergence requirement
        maxiters      maximum number of iterations
        monitor       if non-zero, progress info is output to stdout

        OUTPUTS
        -------
        an array containing the final values
        lowest value of the error function
        number of iterations taken to get here
        """

        iter = 0

        for iter in range(0, maxiters):
            # Identify highest and lowest vertices

            self.highest = 0
            self.lowest = 0
            for vertex in range(1, self.numvars + 1):
                if self.errors[vertex] > self.errors[self.highest]:
                    self.highest = vertex
                if self.errors[vertex] < self.errors[self.lowest]:
                    self.lowest = vertex

            # Identify second-highest vertex

            self.secondhighest = self.lowest
            for vertex in range(0, self.numvars + 1):
                if vertex == self.highest:
                    continue
                elif vertex == self.secondhighest:
                    continue
                elif self.errors[vertex] > self.errors[self.secondhighest]:
                    self.secondhighest = vertex

            # Test for convergence:
            #   compute the average merit figure (ugly)

            S = 0.0
            for vertex in range(0, self.numvars + 1):
                S = S + self.errors[vertex]
            F2 = S / (self.numvars + 1)

            #   compute the std deviation of the merit figures (ugly)

            S1 = 0.0
            for vertex in range(0, self.numvars + 1):
                S1 = S1 + (self.errors[vertex] - F2)**2
            T = math.sqrt(S1 / self.numvars)

            # Optionally, print progress information

            if monitor:
                print('\r' + 72 * ' ', end=' ')
                print('\rIteration = %d   Best = %f   Worst = %f' % \
                      (iter,self.errors[self.lowest],self.errors[self.highest]), end=' ')
                sys.stdout.flush()

            if T <= epsilon:
                # We converged!  Break out of loop!

                break
            else:
                # Didn't converge.  Keep crunching.

                # Calculate centroid of simplex, excluding highest vertex
                # store centroid in element N+1

                # loop over coordinates
                for x in range(0, self.numvars):
                    S = 0.0
                    for vertex in range(0, self.numvars + 1):
                        if vertex == self.highest:
                            continue
                        S = S + self.simplex[vertex][x]
                    self.simplex[self.numvars + 1][x] = S / self.numvars

                # reflect the simplex across the centroid
                # store reflected point in elem. N + 2 (and self.guess)

                self.reflect_simplex()
                self.currenterror = self.testfunc(self.guess)

                if self.currenterror < self.errors[self.highest]:
                    self.accept_reflected_point()

                if self.currenterror <= self.errors[self.lowest]:
                    self.expand_simplex()
                    self.currenterror = self.testfunc(self.guess)

                    # at this point we can assume that the highest
                    # value has already been replaced once
                    if self.currenterror < self.errors[self.highest]:
                        self.accept_expanded_point()
                elif self.currenterror >= self.errors[self.secondhighest]:
                    # worse than the second-highest, so look for
                    # intermediate lower point

                    self.contract_simplex()
                    self.currenterror = self.testfunc(self.guess)

                    if self.currenterror < self.errors[self.highest]:
                        self.accept_contracted_point()
                    else:
                        self.multiple_contract_simplex()

        # Either converged or reached the maximum number of iterations.
        # Return the lowest vertex and the currenterror.

        for x in range(0, self.numvars):
            self.guess[x] = self.simplex[self.lowest][x]
        self.currenterror = self.errors[self.lowest]
        return self.guess, self.currenterror, iter

    # same as expand, but with alpha < 1; kC = 0.5 fine with NR

    def contract_simplex(self):
        for x in range(0, self.numvars):
            self.guess[x] = self.kC * self.simplex[self.highest][x] + (1 - self.kC) * self.simplex[self.numvars + 1][x]
        return

    # expand: if P is vertex and Q is centroid, alpha-expansion is Q + alpha*(P-Q),
    #         or (1 - alpha)*Q + alpha*P; default alpha is 2.0; agrees with NR
    def expand_simplex(self):
        for x in range(0, self.numvars):
            self.guess[x] = self.kE * self.guess[x]                 + (1 - self.kE) * self.simplex[self.numvars + 1][x]
        return

    # reflect: if P is vertex and Q is centroid, reflection is Q + (Q-P) = 2Q - P,
    #          which is achieved for kR = -1 (default value); agrees with NR
    def reflect_simplex(self):
        # loop over variables
        for x in range(0, self.numvars):
            self.guess[x] = self.kR * self.simplex[self.highest][x] + (1 - self.kR) * self.simplex[self.numvars + 1][x]
            # store reflected point in elem. N + 2
            self.simplex[self.numvars + 2][x] = self.guess[x]
        return

    # multiple contraction: around the lowest point; agrees with NR

    def multiple_contract_simplex(self):
        for vertex in range(0, self.numvars + 1):
            if vertex == self.lowest:
                continue
            for x in range(0, self.numvars):
                self.simplex[vertex][x] = 0.5 * (self.simplex[vertex][x] + self.simplex[self.lowest][x])
        self.calculate_errors_at_vertices()
        return

    def accept_contracted_point(self):
        self.errors[self.highest] = self.currenterror
        for x in range(0, self.numvars):
            self.simplex[self.highest][x] = self.guess[x]
        return

    def accept_expanded_point(self):
        self.errors[self.highest] = self.currenterror
        for x in range(0, self.numvars):
            self.simplex[self.highest][x] = self.guess[x]
        return

    def accept_reflected_point(self):
        self.errors[self.highest] = self.currenterror
        for x in range(0, self.numvars):
            self.simplex[self.highest][x] = self.simplex[self.numvars + 2][x]
        return

    def calculate_errors_at_vertices(self):
        for vertex in range(0, self.numvars + 1):
            if vertex == self.lowest:
                # compute the error unless we're the lowest vertex
                continue
            for x in range(0, self.numvars):
                self.guess[x] = self.simplex[vertex][x]
            self.currenterror = self.testfunc(self.guess)
            self.errors[vertex] = self.currenterror
        return

def myfunc(args):
    return abs(args[0] * args[0] * args[0] * 5 - args[1] * args[1] * 7 + math.sqrt(abs(args[0])) - 118)

def main():
    s = Simplex(myfunc, [1, 1, 1], [2, 4, 6])
    values, err, iter = s.minimize()
    print('args = ', values)
    print('error = ', err)
    print('iterations = ', iter)

if __name__ == '__main__':
    main()




# Get Strain/Stress from ImageD11 UBI/map files
# Copyright (C) 2015 Younes ELHACHI
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


"""
Tkinter wrapper for ImageD11 indexing object

All communication should be via parent guicommander object

Owner of the plot3d window
"""

from .listdialog import listdialog


class guisolver:

    def __init__(self,parent):
        """
        Parent (arg) is a hook to features of the parent gui
        sets up eps_sig_solver menuitems
        """
        self.parent=parent

        self.menuitems = ( "Strain/Stress", 0,
                           [ ( "Load ubis", 0, self.loadubis),
                             ( "Load parameters", 1, self.loadfileparameters),
                             ( "Edit parameters", 0, self.editparameters),
                             ( "Compute and save strain and stress", 0, self.compute_save_epsig),
                             ( "Save parameters", 0, self.saveparameters)
                           ] )


    def loadubis(self):
        """ see eps_sig_solver.loadmap """
        filename=self.parent.opener.show(
            title="File containing ubi matrices",
            filetypes=[ ("Grain map files", "*.map"),
                        ("UBI files", "*.ubi") ] )
        self.parent.guicommander.execute("solver","loadmap",filename)

    def compute_save_epsig(self):
        """ see eps_sig_solver.compute_eps_sig """
        filename=self.parent.saver.show(title="File to save strain and stress")
        self.parent.guicommander.execute("solver","compute_write_eps_sig",filename)

    def loadfileparameters(self):
        """ see eps_sig_solver.loadpars and parameters.loadpars """
        filename=self.parent.opener.show(
            title="File containing unit cell and elastic constants",
            filetypes = [ ("Parameter files", "*.prm"),
                          ("Parameter files", "*.par") ] )
        self.parent.guicommander.execute("solver","loadpars",filename)

    def saveparameters(self):
        """ see eps_sig_solver.savepars and parameters.savepars """
        filename=self.parent.saver.show(title="File to save unit cell and elastic constants")
        self.parent.guicommander.execute("solver","savepars",filename)


    def editparameters(self):
        """
        Has the eps_sig_solver object update its parameter object
               eg : savepars(None)
        gets a copy of the parameter object
        Allows user to edit parameters
        Has the eps_sig_solver object update itself from the parameter object
               eg : loadpars(None)
        """
        # First make the eps_sig_solver update its parameters object
        self.parent.guicommander.execute("solver","updateparameters") # no filename arg
        # Now borrow a copy to read them and edit
        pars = self.parent.guicommander.getdata("solver","pars")
        d=listdialog(self.parent,items=pars,title="unit cell and elastic constants")
        self.parent.guicommander.execute("solver","parameterobj.set_parameters",d.result) 
        self.parent.guicommander.execute("solver","loadpars") # and use them 
        



import unittest

import numpy as np

from ImageD11 import cImageD11


class test_connectedpixels(unittest.TestCase):
    def setUp(self):
        dims = (10,10)
        dark = np.ones(dims)
        flood = np.ones(dims)

    def test_1(self):
        for t in [np.uint8, np.int8, np.uint16, np.int16,
                  np.int32, np.uint32, np.float32, np.float]:
            data = np.array( [[ 0, 0, 0, 0, 0, 0, 0],
                             [ 0, 0, 0, 1, 0, 0, 0],
                             [ 0, 0, 0, 0, 0, 0, 0],
                             [ 0, 0, 0, 0, 0, 0, 0]],t)
            bl = np.zeros(data.shape,np.int32)
            cImageD11.connectedpixels(
                data.astype(np.float32),
                bl,
                0.1)  # threshold
            err = np.sum(np.ravel(data-bl))
            self.assertEqual(err, 0)

    def test_1_shape(self):
        for t in [np.uint8, np.int8, np.uint16, np.int16,
                  np.int32, np.uint32, np.float32, np.float]:
            data = np.array( [[ 1, 0, 1, 0, 1, 0, 1],
                             [ 1, 0, 1, 0, 1, 0, 1],
                             [ 1, 0, 1, 0, 0, 1, 0],
                             [ 1, 1, 1, 1, 1, 0, 1]],t)
            bl = np.zeros(data.shape,np.int32)
            cImageD11.connectedpixels(
                data.astype(np.float32),
                bl,
                0.1)  # threshold
            err = np.sum(np.ravel(data-bl))
            self.assertEqual(err, 0)

    def test_2_shapes(self):
        for t in [np.uint8, np.int8, np.uint16, np.int16,
                  np.int32, np.uint32, np.float32, np.float]:
            data = np.array( [[ 1, 0, 1, 0, 2, 0, 2],
                             [ 1, 0, 1, 0, 2, 0, 2],
                             [ 1, 0, 1, 0, 0, 2, 0],
                             [ 1, 1, 1, 0, 2, 0, 2]],t)
            bl = np.zeros(data.shape,np.int32)
            cImageD11.connectedpixels(
                data.astype(np.float32),
                bl,
                0.1)  # threshold
            err = np.sum(np.ravel(data-bl))
            self.assertEqual(err, 0)




class test_blobproperties(unittest.TestCase):

    def test_prop_names(self):
        names="""  s_1,       /* 1 Npix */
  s_I,       /* 2 Sum intensity */
  s_I2,      /* 3 Sum intensity^2 */
  s_fI,      /* 4 Sum f * intensity */
  s_ffI,     /* 5 Sum f * f* intensity */
  s_sI,      /* 6 Sum s * intensity */
  s_ssI,     /* 7 Sum s * s * intensity */
  s_sfI,     /* 8 Sum f * s * intensity */
  s_oI,         /* 9 sum omega * intensity */
  s_ooI,        /*  */
  s_soI,        /* 10 sum omega * s * intensity */
  s_foI,        /* 11 sum omega * f * intensity */
  mx_I,      /* 12  Max intensity */
  mx_I_f,    /* 13 fast at Max intensity */
  mx_I_s,    /* 14 slow at Max intensity */
  mx_I_o,    /* 15 omega at max I */
  bb_mx_f,      /* 16 max of f */
  bb_mx_s,      /* 17 max of s */
  bb_mx_o,      /* 18 max of omega */
  bb_mn_f,      /* 19 min of f */
  bb_mn_s,      /* 20 min of s */
  bb_mn_o,      /* 22 min of o */
  avg_i,   /* Average intensity */
  f_raw,       /* fast centre of mass */
  s_raw,       /* slow centre of mass */
  o_raw,       /* omega centre of mass */
  m_ss,   /* moments */
  m_ff,
  m_oo,
  m_sf,
  m_so,
  m_fo,
  f_cen,  /* Filled in elsewhere  - spatial correction */
  s_cen,  /* ditto */
  dety, /*Filled in elsewhere  - flip to HFPO book */
  detz, /*Filled in elsewhere  - flip to HFPO book */
  NPROPERTY ,   /* Number of properties if starting at 0 */ """
        nl = names.split("\n")
        namelist = [n.split(",")[0].lstrip().rstrip()
                    for n in nl]
        i = 0
        while i < len(namelist):
            self.assertEqual(i,getattr(cImageD11,namelist[i]))
            i += 1

    def test_find_max(self):
        for t in [np.uint8, np.int8, np.uint16, np.int16,
                  np.int32, np.uint32, np.float32, np.float]:
            data = np.array( [[ 1, 0, 1],
                             [ 1, 0, 1],
                             [ 1, 8, 1],
                             [ 1, 1, 1]],t)
            bl = np.zeros(data.shape,np.int32)
            npx = cImageD11.connectedpixels(
                data.astype(np.float32),bl,0.1)
            self.assertEqual(npx,1)
            err = np.sum(np.ravel(data-bl))
            self.assertEqual(err, 7) # 8-1
            res = cImageD11.blobproperties(data.astype(np.float32),
                                           bl,
                                           npx,
                                           omega=22.)
            from ImageD11.cImageD11 import s_1, s_I, s_I2, \
                s_fI, s_ffI, s_sI, s_ssI, s_sfI, \
                bb_mn_f, bb_mn_s, bb_mx_f, bb_mx_s,\
                bb_mn_o, bb_mx_o, \
                mx_I, mx_I_f, mx_I_s, mx_I_o
            #            print res,res.shape
            self.assertAlmostEqual(res[0][s_1],10)
            self.assertAlmostEqual(res[0][mx_I],8)
            self.assertAlmostEqual(res[0][mx_I_f],1)
            self.assertAlmostEqual(res[0][mx_I_s],2)
            self.assertAlmostEqual(res[0][mx_I_o],22)




class testbloboverlaps(unittest.TestCase):
    def test1(self):
        import sys
        data1 =  np.array([[ 1, 0, 1, 0, 0, 0, 0],
                          [ 1, 0, 1, 0, 0, 0, 0],
                          [ 1, 0, 1, 1, 0, 0, 0],
                           [ 1, 1, 1, 0, 0, 0, 0]],np.float32)
        bl1 = np.zeros(data1.shape, np.int32)
        np1 = cImageD11.connectedpixels(data1,bl1,0.1)
        data2 =  np.array([[ 0, 0, 0, 0, 2, 0, 2],
                          [ 0, 0, 0, 0, 2, 0, 2],
                          [ 0, 0, 0, 2, 0, 2, 0],
                           [ 0, 0, 0, 0, 2, 0, 2]],np.float32)
        bl2 = np.zeros(data2.shape, np.int32)
        np2 = cImageD11.connectedpixels(data2,bl2,0.1)
        r1 = cImageD11.blobproperties(data1, bl1, np1, omega=-10.0)
        r2 = cImageD11.blobproperties(data2, bl2, np2, omega=10.0)

        cImageD11.bloboverlaps(bl1,np1,r1,
                                     bl2,np2,r2, verbose=0)
        # check r1 is zeroed
        err = np.sum(np.ravel(r1))
        self.assertAlmostEqual(err,0.,6)
        from ImageD11.cImageD11 import s_1, s_I, s_I2, \
            s_fI, s_ffI, s_sI, s_ssI, s_sfI, \
            bb_mn_f, bb_mn_s, bb_mx_f, bb_mx_s,\
            bb_mn_o, bb_mx_o,\
            mx_I, mx_I_f, mx_I_s
        # check total pixels
        self.assertAlmostEqual(r2[0,s_1], 18.0, 6)
        self.assertAlmostEqual(r2[0,s_I], 26.0, 6)
        self.assertAlmostEqual(r2[0,bb_mn_f], 0.0, 6)
        self.assertAlmostEqual(r2[0,bb_mx_f], 6.0, 6)
        self.assertAlmostEqual(r2[0,bb_mn_s], 0.0, 6)
        self.assertAlmostEqual(r2[0,bb_mx_s], 3.0, 6)
        self.assertAlmostEqual(r2[0,bb_mn_o], -10.0, 6)
        self.assertAlmostEqual(r2[0,bb_mx_o],  10.0, 6)

    def test2(self):
        import sys
        data1 =  np.array([[ 1, 0, 0, 0, 0, 1, 1],
                          [ 1, 0, 2, 0, 0, 1, 1],
                          [ 1, 0, 2, 2, 0, 0, 0],
                           [ 0, 0, 2, 0, 0, 0, 0]],np.float32)
        bl1 = np.zeros(data1.shape, np.int32)
        np1 = cImageD11.connectedpixels(data1,bl1,0.1)
        data2 =  np.array([[ 0, 0, 0, 0, 2, 0, 0],
                          [ 0, 0, 0, 0, 2, 2, 0],
                          [ 0, 0, 0, 2, 0, 0, 0],
                           [ 0, 0, 0, 0, 0, 0, 0]],np.float32)
        bl2 = np.zeros(data2.shape,np.int32)
        np2 = cImageD11.connectedpixels(data2,bl2,0.1)
        r1 = cImageD11.blobproperties(data1, bl1, np1)
        r2 = cImageD11.blobproperties(data2, bl2, np2)

        cImageD11.bloboverlaps(bl1,np1,r1,
                                     bl2,np2,r2, verbose=0)
        # check which peaks are zeroed
        self.assertAlmostEqual(np.sum(r1[1]),0.,6)
        self.assertAlmostEqual(np.sum(r1[2]),0.,6)

        from ImageD11.cImageD11 import s_1, s_I, s_I2, \
            s_fI, s_ffI, s_sI, s_ssI, s_sfI, \
            bb_mn_f, bb_mn_s, bb_mx_f, bb_mx_s,\
            bb_mn_o, bb_mn_f,\
            mx_I, mx_I_f, mx_I_s
        self.assertAlmostEqual(r1[0,s_1], 3, 6)
        self.assertAlmostEqual(r1[0,s_I], 3, 6)

        # check total pixels
        self.assertAlmostEqual(r2[0,s_1], 12.0, 6)
        self.assertAlmostEqual(r2[0,s_I], 20.0, 6)

    def test3(self):
        import sys
        data1 =  np.array([[ 0, 0, 0, 0, 0, 0, 0],
                          [ 0, 0, 0, 0, 0, 0, 0],
                          [ 0, 1, 1, 1, 1, 1, 0],
                           [ 0, 0, 0, 0, 0, 0, 0]],np.float32)
        bl1 = np.zeros(data1.shape,np.int32)
        np1 = cImageD11.connectedpixels(data1,bl1,0.1)
        data2 =  np.array([[ 0, 0, 0, 0, 0, 0, 0],
                          [ 0, 2, 0, 0, 0, 2, 0],
                          [ 0, 2, 0, 0, 0, 2, 0],
                           [ 0, 2, 0, 0, 0, 2, 0]],np.float32)
        bl2 = np.zeros(data2.shape,np.int32)
        np2 = cImageD11.connectedpixels(data2,bl2,0.1)
        r1 = cImageD11.blobproperties(data1, bl1, np1)
        r2 = cImageD11.blobproperties(data2, bl2, np2)

        cImageD11.bloboverlaps(bl1,np1,r1,
                                     bl2,np2,r2, verbose=0)

        # check which peaks are zeroed
        self.assertAlmostEqual(np.sum(r1[0]),0.,6)
        # Results pile up all in r2[0]
        self.assertAlmostEqual(np.sum(r2[1]),0.,6)
        from ImageD11.cImageD11 import s_1, s_I, s_I2, \
            s_fI, s_ffI, s_sI, s_ssI, s_sfI, \
            bb_mn_f, bb_mn_s, bb_mx_f, bb_mx_s,\
            mx_I, mx_I_f, mx_I_s
        self.assertAlmostEqual(r2[0,s_1],11, 6)
        self.assertAlmostEqual(r2[0,s_I],17, 6)





if __name__=="__main__":
    unittest.main()






from ImageD11 import unitcell, parameters, transform
import numpy as np
import matplotlib.pylab as plt
from mpl_toolkits.mplot3d import Axes3D
import sys

pars = parameters.read_par_file( sys.argv[1] )

# Input unit cell from par file
u = unitcell.unitcell_from_parameters(pars)

# Generate diffraction spots for unit cell:
u.makerings(0.7)
hkls = []
for d in u.ringds:
    hkls += u.ringhkls[d]
hkls = np.array(hkls).T

# A grain orientation
U0 = transform.detector_rotation_matrix(0.1, 0.2, 0.3 )
UB0 = np.dot( U0, u.B)


# Ideal diffraction vectors:
gve = np.dot( UB0, hkls )
modg= np.sqrt((gve*gve).sum(axis=0))

# print scattering vectors in order
order = np.argsort(modg)
for i in order[10::-1]:
    print i,hkls.T[i],modg[i],gve.T[i]


def make_uniq( UBIlist ):
    # consider two UBI as equal if dot(UBIa, inv(UBIb)) == {1|0}
    # Sort by trace before to get back the ones with biggest traces
    dsu = [ (np.trace(x)+i*1e-10, x) for i,x in enumerate(UBIlist) ]
    dsu.sort()
    UL = [x[1] for x in dsu[::-1]]
    uniqs = [( UL[0] , np.linalg.inv( UL[0] ) ), ]
    for u1 in UL[1:]:
        new = True
        for u2,u2i in uniqs:
            testmat = np.dot( u1, u2i )
            itestmat = np.round( testmat )
            scor = abs(testmat - itestmat).sum()
            if scor < 1e-8:
                # Seen already
                new = False
                break
        if new:
            uniqs.append( (u1, np.linalg.inv( u1 ) ) )
    return [x[0] for x in uniqs]

def get_ubis( h1, h2, UB0, u ):
    # Compute g-vectors for these peaks in current unit cell
#    print "Using h1",h1,"h2",h2 
    g1 = np.dot( UB0, h1 )
    g2 = np.dot( UB0, h2 )
    modg1 = np.sqrt( np.dot( g1, g1 ) )
    modg2 = np.sqrt( np.dot( g2, g2 ) )
    # find unitcell powder rings for these
    r1 = np.argmin( abs(u.ringds-modg1) )
    r2 = np.argmin( abs(u.ringds-modg2) )
    # Generate orientations for these:
    u.orient( r1, g1, r2, g2)
#    print "Found",len(u.UBIlist),"orientations" 
    return u.UBIlist


# peak pair to use for twinning:
UBIs=[]
for h1, h2 in [ ( (1,0,0), (0,1, 1) ),
                ( (1,0,0), (0,1,-1) ),
                ( (0,1,0), (1,0, 1) ),
                ( (0,1,0), (1,0,-1) ),
                ( (0,0,1), (1,1, 0) ),
                ( (0,0,1),(-1,1, 0) ) ]:
        UBIs+=[x for x in get_ubis( h1, h2, UB0, u )]
uniqs = make_uniq( UBIs)

# Now do that again for each uniq one:
UBIs=[x for x in uniqs]

for UBi1 in uniqs:
#    print UBi1
    for h1, h2 in [ ( (1,0,0), (0,1, 1) ),
                ( (1,0,0), (0,1,-1) ),
                ( (0,1,0), (1,0, 1) ),
                ( (0,1,0), (1,0,-1) ),
                ( (0,0,1), (1,1, 0) ),
                ( (0,0,1),(-1,1, 0) ) ]:
        UB1 = np.linalg.inv( UBi1 )
        UBIs+=[x for x in get_ubis( h1, h2, UB1, u )]

print len(uniqs)
uniqs2 = make_uniq( UBIs)
print len(uniqs2)

# Now do that again for each uniq one:
UBIs=[x for x in uniqs2]

for UBi1 in uniqs2:
#    print UBi1
    for h1, h2 in [ ( (1,0,0), (0,1, 1) ),
                ( (1,0,0), (0,1,-1) ),
                ( (0,1,0), (1,0, 1) ),
                ( (0,1,0), (1,0,-1) ),
                ( (0,0,1), (1,1, 0) ),
                ( (0,0,1),(-1,1, 0) ) ]:
        UB1 = np.linalg.inv( UBi1 )
        UBIs+=[x for x in get_ubis( h1, h2, UB1, u )]
uniqs3 = make_uniq( UBIs)
print len(uniqs3)


# Now do that again for each uniq one:
UBIs=[x for x in uniqs2]

for UBi1 in uniqs3:
#    print UBi1
    for h1, h2 in [ ( (1,0,0), (0,1, 1) ),
                ( (1,0,0), (0,1,-1) ),
                ( (0,1,0), (1,0, 1) ),
                ( (0,1,0), (1,0,-1) ),
                ( (0,0,1), (1,1, 0) ),
                ( (0,0,1),(-1,1, 0) ) ]:
        UB1 = np.linalg.inv( UBi1 )
        UBIs+=[x for x in get_ubis( h1, h2, UB1, u )]
uniqs4 = make_uniq( UBIs)
print len(uniqs4)


uniqs= uniqs4

# Plot out peak positions on detector for the uniqs
fig = plt.figure()
ax1 = fig.add_subplot(221)
ax2 = fig.add_subplot(223)
ax3 = fig.add_subplot(222, projection='3d')
fig2= plt.figure(2)
ax4 = fig2.add_subplot(111)
colors = 'bgryck'

hkls = np.array([(0,0,2),(0,0,-2),(2,0,0),(-2,0,0),(0,2,0),(0,-2,0)]).T
XX=True
def doplot(ubi, col):
    g = np.dot( np.linalg.inv( ubi ), hkls )
    tth, (eta0,eta1), (omega0,omega1) = transform.uncompute_g_vectors( g, pars.get('wavelength') )
    ax1.plot(tth, eta0, col)
    ax1.plot(tth, eta1, col)
    sc, fc = transform.compute_xyz_from_tth_eta( tth, eta0, omega0, **pars.parameters )
    ax2.plot( sc, fc, col)
    sc, fc = transform.compute_xyz_from_tth_eta( tth, eta1, omega1, **pars.parameters )
    ax2.plot( sc, fc, col)
    ax3.plot( g[0], g[1], g[2], col)
    ax4.scatter( eta0, omega0, c=tth)
    cax = ax4.scatter( eta1, omega1, c=tth)
    global XX
    if XX:
        plt.colorbar( cax )
        XX=False
for i,ubi in enumerate(uniqs[1:]):
    col = "%s+"%(colors[i%len(colors)])
    print i,col
    print ubi
    doplot( ubi, col )

doplot( uniqs[0], "ko" )

plt.show()







#!/usr/bin/env python

from __future__ import print_function
try:
    from Tkinter import *
    from tkMessageBox import showinfo
    import tkSimpleDialog as simpledialog
    import tkFileDialog as filedialog
    import Dialog as dialog
except:
    from tkinter import *
    import tkinter.simpledialog as simpledialog
    import tkinter.filedialog as filedialog
    from tkinter.messagebox import showinfo
    import tkinter.dialog as dialog

from PIL import ImageTk,Image

import os,sys

import numpy as np

"""
TO DO

Scroll with the keyboard arrows
Zoom in and out with Page up and Page down (or others)
Draw the box via the keyboard
Implement a status bar and cursor
"""


def preprocess(image, xxx_todo_changeme,maxi=None,mini=None):
    """ Makes data into 8 bit array for plotting """
    (scalex,scaley) = xxx_todo_changeme
    MAXW=4096
    MAXH=4096
    MINSIZE=1
    DEFAULT_HEIGHT=100
    MAXSIZE=max(MAXW,MAXH)
    assert len(image.shape) in (1, 2) or \
           len(image.shape) == 3 and image.shape[2] == 3, \
           "image not correct format"
    if maxi==None:
        themin = float(np.minimum.reduce(np.ravel(image)))
        themax = float(np.maximum.reduce(np.ravel(image)))
    else:
        themin = float(mini)
        themax = float(maxi)
    if len(image.shape) == 1:
        len_x = image.shape[0]
        ys = ((image - themin)/(themax-themin)*(DEFAULT_HEIGHT-1)).astype('B')
        image = (zeros((DEFAULT_HEIGHT, len_x))+255).astype('B')
        for x in range(len_x):
            image[DEFAULT_HEIGHT-1-ys[x],len_x-x-1] = 0
        image = np.transpose(image)
    elif image.dtype.char != 'b':
        try:
            pass  ## image.savespace(0)
            image = 255 * (image - themin) / (themax-themin)
        except:
            print("Exception",themax,themin)
        image = np.where(image<256,image,255)
        image = np.where(image>0,image,0).astype('B')

    len_x, len_y = image.shape[:2]
#    print "Image dimensions...  x=",len_x,"  y=",len_y
    if scalex < MINSIZE:
        scalex = MINSIZE
    if scaley < MINSIZE:
        scaley = MINSIZE
    if scalex > MAXSIZE:
        scalex = MAXSIZE
    if scaley > MAXSIZE:
        scaley = MAXSIZE
    return image, (int(scalex), int(scaley))


def NumerictoImage( data, xxx_todo_changeme1,maxi=None,mini=None):
    (scalex,scaley) = xxx_todo_changeme1
    image, (scalesx,scalesy) = preprocess(data, (scalex,scaley),maxi,mini)
    width, height = image.shape[:2]
    if len(image.shape) == 3:
        mode = rawmode = 'RGB'
        bits = np.transpose(image, (1,0,2)).tostring()
    else:
        mode = rawmode = 'L'
        bits = np.transpose(image, (1,0)).tostring()
    image2 = Image.frombytes(mode, (width, height),
                                      bits, "raw", rawmode)
    image2=image2.resize((int(scalesx),int(scalesy)))
    return image2,(scalesx,scalesy)






class rubber(Frame):

    def __init__(self, datafile=None, bkgfile=None, master=None):
        self.master=master
        Frame.__init__(self, master)
        if datafile==None:
            datafile=filedialog.askopenfilename(initialdir=os.getcwd())
        if type(datafile)==type("string"):
            self.datafile = datafile
            dataobj=openimage(datafile)
            self.data=dataobj.data.astype(np.int)
            try:
                self.omega=float(dataobj.header["Omega"])
            except:
                self.omega = None
        else:
            if type(datafile)==type(np.array([10,11,12])):
                self.data=datafile
                self.datafile="unknown0000.edf"
        pass  ## self.data.savespace(0)
        if type(bkgfile)==type("string"):
            self.bkgfile=bkgfile
            bkgobj=openimage(bkgfile)
            self.bkg=bkgobj.data.astype(np.int)
            self.data=self.data.astype(np.int)-self.bkg
            print("Got your background from",bkgfile,self.bkg.shape)
        else:
            self.bkg = None
        Pack.config(self,expand=1,fill=BOTH)
        self.b=Frame(self)
        Button(self.b,text="Quit",command=self.close).pack(side=RIGHT)
        Button(self.b,text="Toggle log scaled",command=self.togglelog).pack(side=RIGHT)
        Button(self.b,text="Set Maximum",command=self.setmax).pack(side=RIGHT)
        Button(self.b,text="Set Minimum",command=self.setmin).pack(side=RIGHT)
        Button(self.b,text="Autorange",command=self.autorange).pack(side=RIGHT)
        Button(self.b,text="ZoomIn",command=self.zoomin).pack(side=LEFT)
        Button(self.b,text="ZoomOut",command=self.zoomout).pack(side=LEFT)
        Button(self.b,text="ShowROI",command=self.showroi).pack(side=LEFT)
        self.b.pack(side=BOTTOM)

        self.b1=Frame(self)
        Button(self.b1,text="Background",command=self.rdbkg).pack(side=LEFT)
        Button(self.b1,text="Rock",command=self.makerockingcurve).pack(side=LEFT)
        Button(self.b1,text="Prev",command=self.prev).pack(side=LEFT)
        Button(self.b1,text="Next",command=self.mynext).pack(side=LEFT)
        Label(self.b1,text="File number=").pack(side=LEFT)
        self.filenum=StringVar()
        Entry(self.b1,textvariable=self.filenum).pack(side=LEFT)
        Button(self.b1,text="Jump",command=self.jump).pack(side=LEFT)
        Button(self.b1,text="Open",command=self.file).pack(side=LEFT)
        Button(self.b1,text="Pars",command=self.rdpars).pack(side=LEFT)
        Button(self.b1,text="UBIS",command=self.rdubis).pack(side=LEFT)
        self.b1.pack(side=BOTTOM)
        try:
            self.filenum.set("%d"%(int(self.datafile[-8:-4])))
        except:
            try:
                self.filenum.set("%d"%(int(self.datafile[-9:-5])))
            except:
                try:
                    self.filenum.set("%d"%(int(self.datafile[-4:])))
                except:
                    self.filenum.set("0")

        self.status=Label(self,text=self.datafile)
        self.status.pack(side=BOTTOM)
        self.statistics=Label(self,text=self.getstats())
        self.statistics.pack(side=BOTTOM)

        self.key=Label(self,bg='green',bd=0)
        self.key.pack(side=BOTTOM)
        self.keyt=Label(self,bd=0)
        self.keyt.pack(side=BOTTOM,fill=BOTH)

        self.xscrollbar=Scrollbar(self,orient=HORIZONTAL)
        self.xscrollbar.pack(side=BOTTOM,fill=X)
        self.yscrollbar=Scrollbar(self,orient=VERTICAL)
        self.yscrollbar.pack(side=RIGHT,fill=Y)

#        print image.size
        self.scale=[self.data.shape[0],self.data.shape[1]]
        self.zoom=1.0
        self.log=0
        self.canvasObject = Canvas(self,width=512,height=512,
              xscrollcommand=self.xscrollbar.set,yscrollcommand=self.yscrollbar.set,
              scrollregion=(0, 0, self.data.shape[0], self.data.shape[1]))
        self.canvasObject.pack(side=LEFT,fill="both",expand=1)
        self.xscrollbar.config(command=self.canvasObject.xview)
        self.yscrollbar.config(command=self.canvasObject.yview)

        # this is a "tagOrId" for the rectangle we draw on the canvas
        self.rubberbandBox = None
        self.piccy=None
        self.startx = None
        self.mini=self.maxi=None
        while self.scale[0]*self.zoom < 256:
            self.zoom*=2.
        while self.scale[0]*self.zoom > 512:
            self.zoom/=2.
        self.autorange()
        self.viewinself()
        step=float(self.maxi-self.mini)/self.scale[0]
        l=[]
        x=float(self.mini)
        while x <= self.maxi:
            l.append(x)
            x+=step
        self.legend=np.transpose(np.array([l],np.float))
        self.legendimage,(scalesx,scalesy)=NumerictoImage(self.legend,(self.scale[0],10))
        self.legendimage=ImageTk.PhotoImage(self.legendimage)
        self.key.configure(image=self.legendimage)
        self.key.update()
        stri="<---%f          %f--->"%(self.mini,self.maxi)
        self.keyt.configure(text=stri)
        self.keyt.update()


        # and the bindings that make it work..
        self.bind_all("<Prior>", lambda e : self.zoomout())
        self.bind_all("<Next>" ,  lambda e : self.zoomin())
        self.bind("q", self.printkey )
        Widget.bind(self.canvasObject, "<Button-1>", self.mouseDown)
        Widget.bind(self.canvasObject, "<Button1-Motion>", self.mouseMotion)
        Widget.bind(self.canvasObject, "<Button1-ButtonRelease>", self.mouseUp)
        Widget.bind(self.canvasObject, "<Button-3>", self.get_hkl)
        self.parameters={}

    def file(self):
        self.datafile=filedialog.askopenfilename(initialdir=os.getcwd())
        self.readdata()


    def rdpars(self,parfile=None):
        if parfile==None:
            parfile=filedialog.askopenfilename(initialdir=os.getcwd())
        from ImageD11 import parameters
        o=parameters.parameters()
        o.loadparameters(parfile)
        self.parameters=o.parameters

    def rdubis(self,ubifile=None):
        from ImageD11 import indexing
        if ubifile==None:
            ubifile=filedialog.askopenfilename(initialdir=os.getcwd())
        self.ubisread = indexing.readubis(ubifile)
        print("Read into self.ubisread",self.ubisread)

    def rdbkg(self, bkgfile=None):
        if bkgfile == None:
            bkgfile=filedialog.askopenfilename(initialdir=os.getcwd())
        self.bkgfile=bkgfile
        bkgobj=openimage(bkgfile)
        self.bkg=bkgobj.data.astype(np.int)
        self.data=self.data.astype(np.int)-self.bkg
        print("Got your background from",bkgfile,self.bkg.shape)

    def jump(self):
        number=int(self.filenum.get())
        self.datafile=self.name( number )
        self.readdata()

    def name(self,num):
        try:
            current=int(self.datafile[-8:-4])
        except:
            current=int(self.datafile[-9:-5])
        print("current is",current)
        return "%s%04d%s"%(self.datafile[:-8],current+num,
                self.datafile[-4:])

    def mynext(self):
        self.datafile=self.name(1)
        self.readdata()

    def prev(self):
        self.datafile=self.name(-1)
        self.readdata()

    def readdata(self):
        dataobj=openimage(self.datafile)
        self.status.config(text=self.datafile)
        self.data=dataobj.data.astype(np.int)
        if self.bkg is not None:
            try:
                self.data=self.data.astype(np.int)-self.bkg
            except:
                print("Failed to subtract bkg",self.bkg.shape,self.data.shape)
        try:
            self.omega=float(dataobj.header["Omega"])
        except:
            self.omega = None
        self.statistics.configure(text=self.getstats())
        try:
            self.filenum.set("%d"%(int(self.datafile[-8:-4])))
        except:
            self.filenum.set("%d"%(int(self.datafile[-9:-5])))
        self.viewinself()

    def printkey(self,event):
        print("Got a key")
        print(event.char)

    def setmax(self):
        try:
            self.maxi=int(simpledialog.askinteger("Maximum data value?","Maximum data value?"))
            self.viewinself()
            stri="<---%f          %f--->"%(self.mini,self.maxi)
            self.keyt.configure(text=stri)
            self.keyt.update()
        except:
            pass
    def setmin(self):
        try:
            self.mini=int(simpledialog.askinteger("Minimum data value?","Minimum data value?"))
            self.viewinself()
            stri="<---%f          %f--->"%(self.mini,self.maxi)
            self.keyt.configure(text=stri)
            self.keyt.update()
        except:
            pass

    def getstats(self):
        # Convert to float to avoid overflow issues
        t=self.data.astype(np.float)
        sumi=np.sum(np.ravel(t))
        sumisq=np.sum(np.ravel(t*t))
        npixels=t.shape[0]*t.shape[1]
        self.average=sumi/npixels
        self.variance=np.sqrt(sumisq/npixels - self.average*self.average)
        self.mindata=np.minimum.reduce(np.ravel(self.data))
        self.maxdata=np.maximum.reduce(np.ravel(self.data))
        return "%d Pixels, Average = %f, Variance = %f, Min = %f, Max = %f"%(
            npixels,self.average,self.variance,self.mindata,self.maxdata)

    def autorange(self):
        newmin=max(np.minimum.reduce(np.ravel(self.data)),self.average-self.variance)
        newmax=min(np.maximum.reduce(np.ravel(self.data)),self.average+self.variance)
        if self.mini!=newmin or self.maxi != newmax:
            self.mini=newmin
            self.maxi=newmax
            self.viewinself()

    def close(self):
        if self.master==None:
            self.destroy()
        else:
            self.master.destroy()


    def zoomin(self):
        if self.scale[0]*2 > 2050:
            print("Zoom limit of ",self.zoom,"reached")
        else:
            self.zoom*=2
            print("Zooming in to scale ",self.zoom,self.zoom*self.data.shape[0])
            self.viewinself()

    def zoomout(self):
        if self.zoom < 0.02:
            print("Zoom out limit of",self.zoom, "reached")
        else:
            self.zoom/=2
            print("Zooming out to scale ",self.zoom)
            self.viewinself()

    def viewinself(self):
        print(self.maxi,self.mini)
        if self.log==0:
            self.image,(self.scale[0],self.scale[1])=NumerictoImage(self.data,self.data.shape,self.maxi,self.mini)
        else:
            self.image,(self.scale[0],self.scale[1])=NumerictoImage(np.log(self.data+1),self.data.shape,self.maxi,self.mini)
        self.scale[0]*=self.zoom
        self.scale[1]*=self.zoom
        self.scale = [int(v) for v in self.scale]
        print("Image size",self.scale)
        self.displayimage = ImageTk.PhotoImage(self.image.resize(self.scale))
        if self.piccy==None:
            self.piccy=self.canvasObject.create_image(2,2,image=self.displayimage,
                                             anchor=NW,tags="picture")
        self.canvasObject.itemconfig(self.piccy,image=self.displayimage)
        self.canvasObject.config(scrollregion=(0,0,self.scale[0],self.scale[1]))
        self.canvasObject.delete(self.rubberbandBox)
        stri="<---%f          %f--->"%(self.mini,self.maxi)
        self.keyt.configure(text=stri)
        self.keyt.update()
        if self.rubberbandBox is not None:
            z=self.zoom
            self.rubberbandBox = self.canvasObject.create_rectangle(
              z*self.startx, z*self.starty, z*self.endx, z*self.endy, outline='green')

    def getbox(self):
        print(self.startx*self.zoom,self.endx*self.zoom)
        print(self.starty*self.zoom,self.endy*self.zoom)
        xl=int(min(self.startx,self.endx)) ; xh = int(max(self.startx,self.endx))
        yl=int(min(self.starty,self.endy)) ; yh = int(max(self.starty,self.endy))
        print("[",xl,":",xh,",",yl,":",yh,"]",self.data.shape)
        if xl!=xh and yl!=yh :
            return self.data[xl:xh,yl:yh].copy()
        else:
            return None

    def showroi(self):
        selection=self.getbox()
        if selection is not None:
            t = Toplevel()
            junk=rubber(selection,master=t)


    def togglelog(self):
        self.keyt.update()
        if self.log==0:
            self.log=1
            self.legendimage,(scalesx,scalesy)=NumerictoImage(np.log(self.legend+1),(self.scale[0],10))
            self.legendimage=ImageTk.PhotoImage(self.legendimage)
            self.key.configure(image=self.legendimage)
            self.key.update()
            stri="<---%f          %f--->"%(self.mini,self.maxi)
            self.keyt.configure(text=stri)
            self.maxi=np.log(self.maxi+1)
            self.mini=np.log(self.mini+1)
            print(self.maxi,self.mini)
            self.viewinself()
        else:
            self.log=0
            self.legendimage,(scalesx,scalesy)=NumerictoImage(self.legend,(self.scale[0],10))
            self.legendimage=ImageTk.PhotoImage(self.legendimage)
            self.key.configure(image=self.legendimage)
            self.key.update()
            stri="<---%f          %f--->"%(self.mini,self.maxi)
            self.keyt.configure(text=stri)
            self.maxi=np.exp(self.maxi)-1
            self.mini=np.exp(self.mini)-1
            print(self.maxi,self.mini)
            self.viewinself()

    def get_hkl(self, event):
        # canvas x and y take the screen coords from the event and translate
        # them into the coordinate system of the canvas object
        xpos = self.canvasObject.canvasx(event.x)/self.zoom
        ypos = self.canvasObject.canvasy(event.y)/self.zoom
        print("xpos,ypos",xpos,ypos)
        from ImageD11 import transform
        tth,eta = transform.compute_tth_eta( np.array([[xpos], [ypos ]]) ,
                                             **self.parameters )
        print("omega:",self.omega,type(self.omega))
        om = np.array([float(self.omega)])
        print("tth,eta,om",tth,eta,om)
        self.gv = transform.compute_g_vectors(tth,eta,om,float(self.parameters['wavelength']), self.parameters['wedge'])
        self.gv = np.transpose(self.gv)
        s=""
        i=0
        for ubi in self.ubisread:
            h=np.dot(ubi,self.gv.T)
            print("i=%d"%(i),"hkl= %.2f %.2f %.2f"%tuple(h))
            i+=1
            s+="grain %3d\n h = %.2f k=%.2f l = %.2f\n"%(i,h[0],h[1],h[2])
        showinfo("Right clicked",
                 "You click at %f %f, tth=%f eta=%f omega=%f\n %s"%(xpos,ypos,tth,eta,om,s))


    def mouseDown(self, event):
        # canvas x and y take the screen coords from the event and translate
        # them into the coordinate system of the canvas object
        self.canvasObject.delete(self.rubberbandBox)
        self.startx = self.canvasObject.canvasx(event.x)/self.zoom
        self.starty = self.canvasObject.canvasy(event.y)/self.zoom

    def mouseMotion(self, event):
        # canvas x and y take the screen coords from the event and translate
        # them into the coordinate system of the canvas object
        x = self.canvasObject.canvasx(event.x)/self.zoom
        y = self.canvasObject.canvasy(event.y)/self.zoom

        if (self.startx != event.x)  and (self.starty != event.y) :
            self.canvasObject.delete(self.rubberbandBox)
            z=self.zoom
            self.rubberbandBox = self.canvasObject.create_rectangle(self.startx*z, self.starty*z, x*z, y*z, outline='green')
            # this flushes the output, making sure that
            # the rectangle makes it to the screen
            # before the next event is handled
            self.update_idletasks()

    def mouseUp(self, event):
        self.endx = self.canvasObject.canvasx(event.x)/self.zoom
        self.endy = self.canvasObject.canvasy(event.y)/self.zoom
        print(self.startx,self.endx,self.starty,self.endy)

    def makerockingcurve(self):
        if self.startx==None:
            pass
        print(self.startx*self.zoom,self.endx*self.zoom)
        print(self.starty*self.zoom,self.endy*self.zoom)
        xl=int(min(self.startx,self.endx)) ; xh = int(max(self.startx,self.endx))
        yl=int(min(self.starty,self.endy)) ; yh = int(max(self.starty,self.endy))
        print("[",xl,":",xh,",",yl,":",yh,"]",self.data.shape)
        if xl!=xh and yl!=yh :
            print("ROI = ",xl,xh,yl,yh)
            stem = simpledialog.askstring("Filename stem?","Filename stem?",initialvalue=self.datafile[:-8])
            first = simpledialog.askinteger("First File to treat","First file to treat")
            last = simpledialog.askinteger("Last File to treat", "Last file to treat")
            extn=".edf"
            s='files %s%04d%s to %s%04d%s\nArea of interest %d %d %d %d'%(stem,first,extn,stem,last,extn,xl,xh,yl,yh)
            print("Making rocking curve for:",s)

            d = dialog.Dialog(None, {'title': 'Make a rocking curve',
                      'text':  "Sorry - this is not working yet!!!" ,
                      'bitmap': dialog.DIALOG_ICON,
                      'default': 0,
                      'strings': ('Do it now please!',
                                  """That's not what I meant!""")})
            return
            print(d.num)
            if d.num==0:
                import rocker
                line=rocker.rock(stem,first,last,extn,xl,xh,yl,yh)
                import twodplotml
                twodplotml.twodplot(parent=Toplevel(),data=line)
                print(line)
            else:
                print("What a waste of time that was!")



import sys
from fabio.openimage import openimage
if len(sys.argv)>2:
    test = rubber(sys.argv[1],bkgfile=sys.argv[2])
elif len(sys.argv)==2:
    test = rubber(sys.argv[1])
else:
    test = rubber()

test.mainloop()



import ctypes, fabio, os, time, numpy as np, pylab as pl

def compile(x):
    if os.path.exists("%s.so"%(x)):
        os.remove("%s.so"%(x))
    os.system("gcc -shared -O3 -fPIC %s.c -o %s.so"%(x,x))
    
x = "../src/sigma_cut"
compile(x)
sc=ctypes.CDLL("%s.so"%(x))
sc.threshold_image.restype = None
sc.threshold_image.argtypes = [
    np.ctypeslib.ndpointer( dtype=np.float32,
                            ndim=1, flags="CONTIGUOUS"),
    ctypes.c_int,
    ctypes.c_float,
    np.ctypeslib.ndpointer( dtype=np.int8,
                            ndim=1, flags="CONTIGUOUS") ]

sc.histogram_image.restype = None
sc.histogram_image.argtypes = [
    np.ctypeslib.ndpointer( dtype=np.float32,
                            ndim=1, flags="CONTIGUOUS"),
    ctypes.c_int,
    ctypes.c_float,
    ctypes.c_float,
    np.ctypeslib.ndpointer( dtype=np.int32,
                            ndim=1, flags="CONTIGUOUS"),
    ctypes.c_int,]



if 1:
    fname = "testoverlaps0000.edf"
    im=fabio.open(fname).data.astype(np.float32)
    hh = np.zeros(256, np.int32)
    low = 24
    step = 2
    

else:
    fname = "/data/id11/nanoscope/Commissioning/2017Feb/gold/Au6_s0_048_b/Au6_s0_048_b0432.edf.gz"
    bname = "/data/id11/nanoscope/Commissioning/2017Feb/gold/pks/bkg.edf"
    im=fabio.open(fname).data.astype(np.float32)
    im = im - fabio.open(bname).data.astype(np.float32)
    h,b = np.histogram( im , bins=np.arange(-128,128))
    bc = (b[1:]+b[:-1])/2
    pl.plot( bc, h, "-")
    hh = np.zeros(256, np.int32)
    low = -50
    step = 1

sc.histogram_image( im.ravel(), len(im.ravel()), low, step, hh, len(hh))    
# Notes:
#
#  The cutoff in the iterative approach depends if we come from below or above
#  ... the process was not converging properly for the testoverlaps image
#
#  Alternative idea:
#    Compute the histogram of the image
#    Use the histogram to determine the cutoff (e.g. fit low side of "peak")
#  TODO : use the histogram...
    

pl.plot( np.arange(low,low+len(hh)*step,step), hh, "-")
pl.semilogy()
pl.show()
mask = np.zeros( im.shape, np.int8 )
start = time.time()
sc.threshold_image( im.ravel(), len(im.ravel()), 4., mask.ravel())
end = time.time()
print(end-start)




from __future__ import print_function

from six.moves import input

# Try fitting using theta/eta/omega instead of x/y/omega or gx/gy/gz

# Each peak comes with sc, fc, omega
#  convert this to tth, eta, omega and dtth/dt, deta/dt

# omega depends on UB (not translations)

# sc,fc depends on UB and translations

# tth depends on B, translations but not U

# eta depends on UB and translations

# variables to fit: U(3), B(6), t(3)
#  B0 B
# write U as Unow + Ushift...


# TODO :
# ... replace numeric derivatives by analytical
# ... reformulate UB as rx,ry,rz,a*,b*,c*,alpha* etc
# ... refactor LSQ to allow symmetry constraints



from ImageD11 import columnfile, grain, parameters, transform, gv_general, indexing
import sys
import numpy as np

TESTING=0

if TESTING:
    import pylab as pl
    pl.ion()

def Dcalc( UB, hkls, wedge, chi, wavelength, etao, omega):
    """
    Finds "C"alculated 2thetac, etac, omegac for a grain
    (Note that these do not depend on the grain position)

    UB = 3x3 UB matrix
    hkls = 3xn list of peaks (normally integers)
    wedge = tilt around y under rotation axis
    chi = tilt around beam under rotation axis
    wavelength = of the radiation, scales UB
    etao = observed azimuth values (angle on powder ring)
    omega = observed(?) sample rotation angles

    ???    ysign determines which solution (there are usually two)
    """
    # computed g-vectors
    g = np.dot( UB  , hkls )
    assert g.dtype == np.float
    if 0:
        print(g.shape)
        print(hkls.shape)
        for i in range(10):
            print(g[:,i], end=' ')
            print(hkls[:,i], end=' ')
            print(np.dot(np.linalg.inv( UB ), g[:,i] ))
    
    # wedge/chi matrix for axis orientation
    wc = gv_general.wedgechi( wedge=wedge, chi=chi )
    assert wc.dtype == np.float
    # computed omega angles for reflections
    omega1, omega2, valid = gv_general.g_to_k(
        g, wavelength,axis=[0,0,-1], pre=None, post=wc )
    assert omega1.dtype == np.float
    # hum - isn't this wc-1 ? 
    cwT = gv_general.chiwedge( wedge=wedge, chi=chi ).T
    # k vectors, scattering in laboratory at angle omega1/2
    k1 = gv_general.k_to_g( g, omega1, axis=[0,0,1],
                            pre = cwT, post=None)
    k2 = gv_general.k_to_g( g, omega2, axis=[0,0,1],
                            pre = cwT, post=None)
    assert k1.dtype == np.float
    # Computed azimuth angles
    #
    # k[1,:] = -ds*c*sin(eta)
    # ------    -------------   .... tan(eta) = -k1/k2
    # k[2,:] =  ds*c*cos(eta)
    #
    eta_one = np.degrees( np.arctan2(-k1[1, :], k1[2, :]) )
    eta_two = np.degrees( np.arctan2(-k2[1, :], k2[2, :]) )
    # 
    # We select the one matching the observed data
    # FIXME - doesn't make sense to do this here
    # pick the solution which is closest to matching in eta/omega
    d1 = angmod(eta_one - etao)
    d2 = angmod(eta_two - etao)
    # Selecting which is the computed value for peaks
    etac   = np.where( abs(d1) < abs(d2), eta_one, eta_two)
    omegac = np.where( abs(d1) < abs(d2), omega1 , omega2 )

    if TESTING > 2:
        err = abs(etac - etao)
        for i in range(len(err)):
            if err[i] < 1:
                continue
            print(i, eta_one[i], eta_two[i], omega1[i], omega2[i])
            print("\t", etao[i], omega[i], etac[i], omegac[i])
        
    # Two theta angles
    mod_g = np.sqrt((g*g).sum(axis=0))
    tthc = np.degrees( 2*np.arcsin(wavelength*mod_g/2) )
    # Return computed values (no derivatives here)
    return  tthc, etac, omegac

def Dobs( t, sc, fc, omega, pars ):
    """
    Compute tth0, eta0 and derivatives given:
    t = (tx,ty,tz) grain position
    sc,fc = pixel positions of spots
    omega = sample rotation angle
    pars = ImageD11 parameters (dict) with wavelength, distance, tilts, etc

    returns tth, eta, dtth, deta
    """
    mypars = pars.parameters.copy()
    mypars['t_x'] = t[0]
    mypars['t_y'] = t[1]
    mypars['t_z'] = t[2]
    # tth, eta for the spots
    tth0, eta0 = transform.compute_tth_eta( (sc, fc), omega=omega, **mypars )
    # arrays for dtth/dt_{xyz} and deta/dt_{xyz}
    dtth=np.zeros( (3,len(sc)) )
    deta=np.zeros( (3,len(sc)) )
    # Step size for numerical derivative
    s = 1.0
    for i,p in enumerate(("t_x","t_y","t_z")):
        # reset
        mypars['t_x'] = t[0]
        mypars['t_y'] = t[1]
        mypars['t_z'] = t[2]
        # shift + and -
        mypars[p] = t[i] + s/2.0
        tth1, eta1 = transform.compute_tth_eta( (sc, fc),
                                                omega=omega, **mypars )
        mypars[p] = t[i] - s/2.0
        tth2, eta2 = transform.compute_tth_eta( (sc, fc),
                                                omega=omega, **mypars )
        dtth[ i ] = (tth1 - tth2)/s
        deta[ i ] = angmod(eta1 - eta2)/s 
    return tth0, eta0, dtth, deta

def angmod( a ):
    """
    Brings angle into range of arctan2 (normally -180, 180)
    """
    ar = np.radians(a)
    return np.degrees( np.arctan2( np.sin( ar ), np.cos( ar ) ) )

class SVDLSQ(object):
    def __init__(self, parnames):
        """
        parnames = List of parameter names for fitting.
        """
        self.parnames = parnames
        self.gradients = {}
        self.differences = []
        self.weights = []
        for name in self.parnames:
            self.gradients[name] = []

    def addobs(self, difference, gradients, weights):
        """
        Add in a series of observations with weights
        Records these ready to solve later

        difference, gradients, weights = block addins
        difference = [nobs]
        gradients = dict[ parnames ][nobs]
        weights = [nobs]
        """
        for name in self.parnames:
            self.gradients[name].append( gradients[name] )
        self.differences.append( difference )
        self.weights.append( weights )

    def solve( self ):
        """
        Solve the least squares problem using an SVD
        TODO ... fill in some residual and error information
        """
        N = len(self.parnames)        # Number of parameters
        M = 0                         # Differences
        for diff in self.differences: # block addins
            M += len(diff)
        # Least squares problem matrix A.x=b
        A = np.zeros( (N, M), np.float )
        b = np.zeros( M, np.float )
        iM = 0
        #print N, M, [len(x) for x in self.differences]
        # FIXME: This loop is doing a numpy.concatenate
        for i, (diff, wt) in enumerate(zip(self.differences, self.weights)):
            b[iM:iM+len(diff)] = diff*wt
            for j, name in enumerate(self.parnames):
                A[j,iM:iM+len(diff)] = self.gradients[name][i]*wt
            iM += len(diff)
        #assert (b == np.concatenate( self.differences )).all()
        # Compute the SVD of A:
        try:
            U, s, V = np.linalg.svd( A, full_matrices = False, compute_uv = True )
        except np.linalg.LinAlgError:
            print("SVD failure!!!")
            raise
        # Clip out small singular values:
        invS = np.where( s > s.max()*1e-9, 1/s, 0 )
        # Pseudo inverse matrix
        self.imat = np.dot(  U*invS, V )
        solution = np.dot( self.imat , b )
        # Save some stuff
        self.U = U
        self.s = s
        self.V = V
        self.errmat = np.dot( U*invS*invS, U.T )
        if TESTING > 0:
            lsqmat = np.dot( A, A.T )
            errmat = np.linalg.inv( lsqmat )
            print(solution)
            print(errmat)
            print(svderrmat)
            assert np.allclose( self.errmat, errmat )
        self.esds = np.sqrt( np.diag( self.errmat ) )
        self.condition = s.max() / s.min() # sqrt ?
        self.sh_esd = solution / self.esds
        # These are shifts
        return solution



def fitone( UB, t, sc, fc, omega, hkls, pars):
    """
    Fit a grain
    """
    npks = len(omega)
    pnames = "t0 t1 t2".split() + ["UB%d%d"%(i,j) for i in range(3) 
                                for j in range(3)]
    S = SVDLSQ( pnames )
    # To do : Allocate LSQ problem gradient array here and keep S in grain (?)
    gradients = {}
    ttho, etao, dtth, deta =  Dobs( t, sc, fc, omega, pars )
    tthc, etac, omegac = Dcalc( UB, hkls,
                                pars.get('wedge'), pars.get('chi'),
                                pars.get('wavelength'),
                                etao, omega)
    # Differences
    errtth = tthc - ttho
    erreta = etac - etao # Why not angmod here ? 
    erromega = angmod( omegac - omega )

    if TESTING>1:
        pl.figure()
        pl.subplot(311)
        pl.plot(etac,errtth, "+")
        pl.subplot(312)
        pl.plot(etac, erromega, "+")
        pl.subplot(313)
        pl.plot(etac, erreta, "+")
        pl.show()
        input()

    
    s = 1e-7
    # FIXME: Gradient arrays (move to Dcalc)
    dtthUB = np.zeros((3,3,npks), np.float)
    detaUB = np.zeros((3,3,npks), np.float)
    domUB = np.zeros((3,3,npks), np.float)
    
    for i in range(3):
        for j in range(3):
            UBc = UB.copy()
            UBc[i,j] -= s/2
            tths1, etas1, omegas1 = Dcalc( UBc, hkls,
                                        pars.get('wedge'), pars.get('chi'),
                                        pars.get('wavelength'),
                                           etao, omega)
            UBc = UB.copy()
            UBc[i,j] += s/2
            tths2, etas2, omegas2 = Dcalc( UBc, hkls,
                                        pars.get('wedge'), pars.get('chi'),
                                        pars.get('wavelength'),
                                           etao, omega)
            dtthUB[i,j] = (tths1 - tths2)
            detaUB[i,j] = (etas1 - etas2)
            domUB[i,j] = (omegas1 - omegas2)
            if TESTING>1:
                pl.subplot(3,3,i*3+j+1)
                pl.plot(detaUB[i,j])
                

    # Three block addins: twotheta, eta, omega:
    for i in range(3):
        gradients["t%d"%(i)] = dtth[i]
        for j in range(3):
            gradients["UB%d%d"%(i,j)] = dtthUB[i,j]
    S.addobs( errtth, gradients, 10. )
    for i in range(3):
        gradients["t%d"%(i)] = deta[i]
        for j in range(3):
            gradients["UB%d%d"%(i,j)] = detaUB[i,j]
    S.addobs( erreta, gradients, 1. )
    for i in range(3):
        gradients["t%d"%(i)] = 0.
        for j in range(3):
            gradients["UB%d%d"%(i,j)] = domUB[i,j]
    S.addobs( erromega, gradients, 1. )
    
    # Find shifts
    sh = S.solve()
    
    # new UB
    UBnew = np.reshape( UB.ravel() + s*sh[-9:], (3,3) )
    
    # returns new translation and new UB
    return t+sh[:3], UBnew, S
    

def refit_makemap( colf, pars, grains ):
    """
    Takes a columnfile, parameters and list of grains
    and repeats the refinement of each grain, keeping
    the same grain assignments
    Returns : new grains
    """ 
    global TESTING
    for i,g in enumerate( grains ):
        d = colf.copy()
        d.filter( d.labels == i )
        # Take peaks belonging to this grain only
        sc = d.sc
        fc = d.fc
        om = d.omega
        hkls = np.array( ( d.h, d.k, d.l ) )
        if i == 1004:
            TESTING=3
        # Get new t, UB, S
        t, UB, S = fitone( g.ub, g.translation, sc, fc, om, hkls, pars)
        for ncycle in range(5):
            t, UB, S = fitone( UB, t, sc, fc, om, hkls, pars)
            if S.sh_esd.max() < 1e-10:
                break
            if TESTING:
                print("translation",t)
                print("Cell",indexing.ubitocellpars(np.linalg.inv(UB)))
        print("Translation %.5f %.5f %.5f"%tuple(t))
        print("Cell %.7f %.7f %.7f %.8f %.8f %.8f"%(indexing.ubitocellpars(np.linalg.inv(UB))))
        g.translation = t
        g.set_ubi(np.linalg.inv( UB ))
    return grains

# TODO:
#
# DONE : Least squares as SVD problem
# - Allow a constraint matrix to be used
# - Compute weights from experimental data (sig/cov etc)
# DONE (but not weighted or printed) : Compute error estimates
# - Allow fitting of other geometrical parameters
# - Allow refinement in terms of other parameterisation of UB (e.g. U, B)
# - Documentation


if __name__=="__main__":
    colfile, parfile, grainsfile, newgrainsfile = sys.argv[1:5]
    c = columnfile.columnfile( colfile )
    p = parameters.read_par_file( parfile )
    g = grain.read_grain_file( grainsfile )    
    grain.write_grain_file( newgrainsfile, refit_makemap( c, p, g) )


    # python teo.py ../test/simul_1000_grains/g187.flt ../test/simul_1000_grains/Al1000/Al1000.par ../test/simul_1000_grains/g187.map teo187fitted.map




from __future__ import print_function, division

import glob, os, multiprocessing, time
import numpy as np
import fabio, h5py
from ImageD11 import sparseframe, cImageD11

def ftomonames( scanpars, folder):
    """ nimages, extn, interlaced, iflip  in scanpars
    Generate the ftomo filename series """
    nums = range(scanpars['nimages'])
    if scanpars['interlaced']:
        stems = [ os.path.join( folder, folder+"0_%04d"+"%s"%(scanpars['extn']) ),
                  os.path.join( folder, folder+"1_%04d"+"%s"%(scanpars['extn']) ) ]
        if scanpars['iflip']:
            s1 = [stems[0]%num for num in nums]
            s2 = [stems[1]%num for num in nums[::-1] ]
        else:
            s1,s2 = [[stem%num for num in nums] for stem in stems]
        # interleave two lists of same length
        return [val for pair in zip(s1,s2) for val in pair]
    else:
        stem = os.path.join( folder, folder+"%04d"+"%s"%(scanpars['extn']) )
        return [stem%num for num in nums]
    
   
def segment( fname, bg, datamem, datapath, block=128, nsigma=3. ):
    """ Does a segmentation of a single file/frame """
    imo = fabio.open(os.path.join( datapath, fname) ) # data
#    print(fname)
    oname = os.path.split( os.path.splitext(fname)[0] )[-1]
    # subtract dark and promote to float32 (without overflow on uint16 I hope)
    cImageD11.uint16_to_float_darksub( datamem.ravel(), bg.ravel(), imo.data.ravel())
    # check offset and noise
    avg,sig = cImageD11.array_mean_var_cut( datamem, cut=nsigma )
    # remove the frelon lines
    datamem.shape = imo.data.shape[0]*(imo.data.shape[1]//block) , block
    cImageD11.frelon_lines( datamem, avg + sig*nsigma )
    datamem.shape = imo.data.shape
    # overwrites datamem (should copy before if you want it
    avg,sig = cImageD11.array_mean_var_cut( datamem, cut=nsigma )
    threshold = avg + sig * nsigma
    # select the pixels to keep:
    rawmask = datamem > threshold
    cleanmask = rawmask.astype(np.int8)
    cImageD11.clean_mask( rawmask.astype(np.int8), cleanmask )
    # save some metadata:
    header = {"filename": fname }
    for k in ("SRCUR","time_of_day"):
        if k in imo.header:
            header[k] = imo.header[k]
    spar = sparseframe.from_data_mask( cleanmask, datamem, header )
    return spar


def write_frames_to_hdf( frames,
                         datapath,
                         fnames,
                         hdfname,
                         hdffolder ):
    """ saves a 1D scan (single ftomo) to a single hdf """
    h = h5py.File( hdfname, "a" )
    g = h.require_group( hdffolder )
    g.attrs['datapath'] = datapath
    for f,name in zip(frames,fnames):
        gname = os.path.split( os.path.splitext(name)[0])[-1]
        f.to_hdf_group( g.require_group( gname ) )
    h.close()


def dofolder( args ):
    """ Processes the files in one folder (single ftomo scan) """
    folder, scanpars = args
    start = time.time()
    fnames = ftomonames( scanpars, folder ) # , nimages, extn, interlaced, iflip )
    hname = os.path.join("hdfs", folder+".hdf")
    bg = fabio.open(
        os.path.join( scanpars['bgfolder'],
                      folder,folder+"_median.edf")
        ).data.astype(np.float32)
    datamem = np.zeros_like( bg )
    spars = [segment(fname, bg, datamem, scanpars['datapath'], block=128)
             for fname in fnames ]
    print("Segmented", folder, time.time()-start)
    write_frames_to_hdf( spars, scanpars['datapath'], fnames, hname, "/" )
    ret = "Wrote %s %.4f /s"%(folder, time.time()-start)
    return ret
 

def harvest_pixels(folders):
    if not os.path.exists("hdfs"):
        os.mkdir( "hdfs" )
    import multiprocessing
    nproc = multiprocessing.cpu_count()//4 # io
    with multiprocessing.Pool( processes = nproc ) as p:
        for x in p.imap_unordered( dofolder, folders ):
            print("Done",x)
        

            
if __name__=="__main__":


    scanpars = {
        "extn" : '.edf',
        "nimages" : 360,
        "interlaced" : True,
        "iflip" : True,
        "datapath" : '/data/id11/nanoscope/Commissioning/2018Apr/difftomo_al_quick',
        "bgfolder" : "/data/id11/jon/difftomo_al_quick/oldmethod",
    }

    folders = [("Al_y%03d_"%(i),scanpars) for i in range(-60,61)]
    
    harvest_pixels( folders )
    






import fabio, sys, numpy as np
stem = sys.argv[1]
if 1:
    out=open("%s.bin"%(stem),"wb")
    for i in range(1024):
        im=fabio.open("%s/%s_%04d.edf"%(stem,stem,i)).data
        im = (im+10)*10
        out.write(im.clip(0,255).astype(np.uint8))

im=fabio.open("%s/%s_%04d.edf"%(stem,stem,0))

open("%s.nrrd"%(stem),"w").write("""NRRD0001
# my first nrrd
type: uchar
dimension: 3
sizes: 1024 1024 1024
encoding: raw
data file: %s.bin
"""%(stem))




from __future__ import print_function


import glob, os, sys, time, fabio

HOME = "/data/visitor/hc1181/id11"

FLOOD= "/data/visitor/hc1181/id11/F4M_irt_flood.edf"

CMD = "huber2bruker.py -n %s -f %d -l %d -F %s -d %s -w 0.158154 -D 8 --omega_zero=-1.5935 --chi_zero=-0.3705 -j 4 --maskfilename=/data/visitor/hc1181/id11/fit2d.msk"

darks = {
    # Key is time in 0.1 s units
    2 : "/data/visitor/hc1181/id11/dark02s.edf",
    5 : "/data/visitor/hc1181/id11/dark05s.edf",
    10 : "/data/visitor/hc1181/id11/dark1s.edf",
    20 : "/data/visitor/hc1181/id11/dark1s.edf",
    30 : "/data/visitor/hc1181/id11/dark3s.edf",
    50 : "/data/visitor/hc1181/id11/dark5s.edf"
    }

def convert(stem,first,last,int_time):
    nearest_int = int(10*float(int_time))
    assert nearest_int in darks
    dark = darks[nearest_int]
    mycmd = CMD%(stem, first, last, FLOOD, dark)
    print(mycmd)
    os.system("%s"%(mycmd))


def bname(s):
    o = fabio.deconstruct_filename(s)
    return "%s_0.%04d"%(o.stem,o.num)

def process(stem):
    print("Checking",stem)
    sys.stdout.flush()
    os.chdir(os.path.join(HOME, stem))
    fl = glob.glob("%s????.edf"%(stem))
    br = glob.glob("%s_0.????"%(stem))
    # edfs = filter(lambda s:s.endswith("edf"), fl)
    # print len(edfs)
    todo = []
    for f in fl:
        newname = bname(f)
        if newname not in br:
            todo.append((newname,f))
    if len(todo) > 0:
        nums = [fabio.getnum(t[0]) for t in todo]
        nums.sort()
    else:
        return
    first=nums[0]
    last = nums[-1]
    int_time = fabio.openimage.openheader(todo[0][1]).header['Integration']
    convert(stem, first, last, int_time )


checked = {0:0}
skip = [30]
skipdir="grid_edna alignKB_edna check_det_edna dark05s_edna 01_SLSCO_006_C1_9__edna".split()

while 1:
    os.chdir(HOME)
    for dirname in glob.glob("*edna"):
        if dirname in skipdir:
            continue
        stem = dirname.replace("_edna","")
        process(stem)
    print("sleeping")
    time.sleep(10)
#    break
